import asyncio
import json
import os
import platform
import shutil
import subprocess
import sys
import tarfile
import zipfile
from pathlib import Path

import aiohttp
import markdown
import tenacity

from randovania import VERSION
from randovania.cli import prime_database
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.interface_common.enum_lib import iterate_enum

_ROOT_FOLDER = Path(__file__).parents[1]
_NINTENDONT_RELEASES_URL = "https://api.github.com/repos/randovania/Nintendont/releases"
zip_folder = "randovania-{}".format(VERSION)


def is_production():
    return os.getenv("PRODUCTION", "false") == "true"


def open_zip(platform_name: str) -> zipfile.ZipFile:
    return zipfile.ZipFile(_ROOT_FOLDER.joinpath(f"dist/{zip_folder}-{platform_name}.zip"),
                           "w", compression=zipfile.ZIP_DEFLATED)


async def get_nintendont_releases(session: aiohttp.ClientSession):
    async with session.get(_NINTENDONT_RELEASES_URL) as response:
        try:
            response.raise_for_status()
            return await response.json()
        except aiohttp.ClientResponseError as e:
            raise RuntimeError("Unable to get Nintendont releases") from e


@tenacity.retry(
    stop=tenacity.stop_after_attempt(5),
    retry=tenacity.retry_if_exception_type(aiohttp.ClientConnectorError),
    wait=tenacity.wait_exponential(multiplier=1, min=4, max=30),
)
async def download_nintendont():
    headers = None
    if "GITHUB_TOKEN" in os.environ:
        headers = {
            "Authorization": f"Bearer {os.environ['GITHUB_TOKEN']}"
        }

    async with aiohttp.ClientSession(headers=headers) as session:
        print("Fetching list of Nintendont releases.")
        releases = await get_nintendont_releases(session)
        latest_release = releases[0]

        download_urls = [
            asset["browser_download_url"]
            for asset in latest_release["assets"]
            if asset["name"] == "boot.dol"
        ]
        if not download_urls:
            raise RuntimeError("No boot.dol found in latest release")

        print(f"Downloading {download_urls[0]}")
        async with session.get(download_urls[0]) as download_response:
            download_response.raise_for_status()
            dol_bytes = await download_response.read()

        final_dol_path = _ROOT_FOLDER.joinpath("randovania", "data", "nintendont", "boot.dol")
        print(f"Saving to {final_dol_path}")
        final_dol_path.write_bytes(dol_bytes)


async def main():
    package_folder = Path("dist", "randovania")
    if package_folder.exists():
        shutil.rmtree(package_folder, ignore_errors=False)

    app_folder = Path("dist", "Randovania.app")
    if app_folder.exists():
        shutil.rmtree(app_folder, ignore_errors=False)

    for game in iterate_enum(RandovaniaGame):
        prime_database.export_as_binary(
            default_data.read_json_then_binary(game)[1],
            _ROOT_FOLDER.joinpath("randovania", "data", "binary_data", f"{game.value}.bin"))

    if is_production():
        server_suffix = "randovania"
    else:
        server_suffix = "randovania-staging"
    configuration = {
        "discord_client_id": 618134325921316864,
        "server_address": f"https://randovania.metroidprime.run/{server_suffix}",
        "socketio_path": f"/{server_suffix}/socket.io",
    }
    with _ROOT_FOLDER.joinpath("randovania", "data", "configuration.json").open("w") as config_release:
        json.dump(configuration, config_release)

    await download_nintendont()

    subprocess.run([sys.executable, "-m", "PyInstaller",
                    "randovania.spec"],
                   check=True)

    if platform.system() == "Windows":
        create_windows_zip(package_folder)
    elif platform.system() == "Darwin":
        create_macos_zip(app_folder)


def create_windows_zip(package_folder):
    if is_production():
        with open_zip("windows") as release_zip:
            for f in package_folder.glob("**/*"):
                print("Adding", f)
                release_zip.write(f, "{}/{}".format(zip_folder, f.relative_to(package_folder)))

            add_readme_to_zip(release_zip)
    else:
        zip_file = os.fspath(_ROOT_FOLDER.joinpath(f"dist/{zip_folder}-windows.7z"))
        subprocess.run(["7z", "a", "-mx=7", "-myx=7", zip_file, os.fspath(package_folder)], check=True)
        subprocess.run(["7z", "rn", zip_file, os.fspath(package_folder), zip_folder], check=True)


def create_macos_zip(folder_to_pack: Path):
    with tarfile.open(_ROOT_FOLDER.joinpath(f"dist/{zip_folder}-macos.tar.gz"), "w:gz") as release_zip:
        release_zip.add(folder_to_pack, f"{zip_folder}/Randovania.app")


def add_readme_to_zip(release_zip):
    with _ROOT_FOLDER.joinpath("README.md").open() as readme_file:
        readme_html = markdown.markdown(readme_file.read())
        release_zip.writestr(zip_folder + "/README.html", readme_html)


if __name__ == '__main__':
    asyncio.run(main())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'racetime_browser_dialog.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_RacetimeBrowserDialog(object):
    def setupUi(self, RacetimeBrowserDialog):
        if not RacetimeBrowserDialog.objectName():
            RacetimeBrowserDialog.setObjectName(u"RacetimeBrowserDialog")
        RacetimeBrowserDialog.resize(528, 453)
        self.verticalLayout = QVBoxLayout(RacetimeBrowserDialog)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.menu_bar = QMenuBar(RacetimeBrowserDialog)
        self.menu_bar.setObjectName(u"menu_bar")

        self.verticalLayout.addWidget(self.menu_bar)

        self.table_widget = QTableWidget(RacetimeBrowserDialog)
        if (self.table_widget.columnCount() < 6):
            self.table_widget.setColumnCount(6)
        __qtablewidgetitem = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(4, __qtablewidgetitem4)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.table_widget.setHorizontalHeaderItem(5, __qtablewidgetitem5)
        self.table_widget.setObjectName(u"table_widget")
        self.table_widget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.table_widget.setTabKeyNavigation(False)
        self.table_widget.setSelectionMode(QAbstractItemView.SingleSelection)
        self.table_widget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.table_widget.setSortingEnabled(True)
        self.table_widget.verticalHeader().setVisible(False)

        self.verticalLayout.addWidget(self.table_widget)

        self.filter_group = QGroupBox(RacetimeBrowserDialog)
        self.filter_group.setObjectName(u"filter_group")
        self.gridLayout = QGridLayout(self.filter_group)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.status_inprogress_check = QCheckBox(self.filter_group)
        self.status_inprogress_check.setObjectName(u"status_inprogress_check")

        self.gridLayout.addWidget(self.status_inprogress_check, 2, 4, 1, 1)

        self.status_cancelled_check = QCheckBox(self.filter_group)
        self.status_cancelled_check.setObjectName(u"status_cancelled_check")

        self.gridLayout.addWidget(self.status_cancelled_check, 2, 7, 1, 1)

        self.filter_status_label = QLabel(self.filter_group)
        self.filter_status_label.setObjectName(u"filter_status_label")

        self.gridLayout.addWidget(self.filter_status_label, 2, 0, 1, 1)

        self.filter_name_label = QLabel(self.filter_group)
        self.filter_name_label.setObjectName(u"filter_name_label")
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.filter_name_label.sizePolicy().hasHeightForWidth())
        self.filter_name_label.setSizePolicy(sizePolicy)

        self.gridLayout.addWidget(self.filter_name_label, 0, 0, 1, 1)

        self.status_finished_check = QCheckBox(self.filter_group)
        self.status_finished_check.setObjectName(u"status_finished_check")

        self.gridLayout.addWidget(self.status_finished_check, 2, 6, 1, 1)

        self.status_invitational_check = QCheckBox(self.filter_group)
        self.status_invitational_check.setObjectName(u"status_invitational_check")

        self.gridLayout.addWidget(self.status_invitational_check, 2, 2, 1, 1)

        self.status_open_check = QCheckBox(self.filter_group)
        self.status_open_check.setObjectName(u"status_open_check")
        self.status_open_check.setChecked(True)

        self.gridLayout.addWidget(self.status_open_check, 2, 1, 1, 1)

        self.filter_name_edit = QLineEdit(self.filter_group)
        self.filter_name_edit.setObjectName(u"filter_name_edit")

        self.gridLayout.addWidget(self.filter_name_edit, 0, 1, 1, 7)

        self.status_pending_check = QCheckBox(self.filter_group)
        self.status_pending_check.setObjectName(u"status_pending_check")

        self.gridLayout.addWidget(self.status_pending_check, 2, 3, 1, 1)


        self.verticalLayout.addWidget(self.filter_group)

        self.label_layout = QHBoxLayout()
        self.label_layout.setSpacing(6)
        self.label_layout.setObjectName(u"label_layout")
        self.status_label = QLabel(RacetimeBrowserDialog)
        self.status_label.setObjectName(u"status_label")
        self.status_label.setWordWrap(True)

        self.label_layout.addWidget(self.status_label)


        self.verticalLayout.addLayout(self.label_layout)

        self.button_box = QDialogButtonBox(RacetimeBrowserDialog)
        self.button_box.setObjectName(u"button_box")
        self.button_box.setMinimumSize(QSize(500, 0))
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.button_box)


        self.retranslateUi(RacetimeBrowserDialog)

        QMetaObject.connectSlotsByName(RacetimeBrowserDialog)
    # setupUi

    def retranslateUi(self, RacetimeBrowserDialog):
        RacetimeBrowserDialog.setWindowTitle(QCoreApplication.translate("RacetimeBrowserDialog", u"Racetime.gg Browser", None))
        ___qtablewidgetitem = self.table_widget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Name", None));
        ___qtablewidgetitem1 = self.table_widget.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Status", None));
        ___qtablewidgetitem2 = self.table_widget.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Entrants", None));
        ___qtablewidgetitem3 = self.table_widget.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Goal", None));
        ___qtablewidgetitem4 = self.table_widget.horizontalHeaderItem(4)
        ___qtablewidgetitem4.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Opened At", None));
        ___qtablewidgetitem5 = self.table_widget.horizontalHeaderItem(5)
        ___qtablewidgetitem5.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Info", None));
        self.filter_group.setTitle(QCoreApplication.translate("RacetimeBrowserDialog", u"Filters", None))
        self.status_inprogress_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"In Progress", None))
        self.status_cancelled_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Cancelled", None))
        self.filter_status_label.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Status", None))
        self.filter_name_label.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Name", None))
        self.status_finished_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Finished", None))
        self.status_invitational_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Invitational", None))
        self.status_open_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Open", None))
        self.status_pending_check.setText(QCoreApplication.translate("RacetimeBrowserDialog", u"Pending", None))
        self.status_label.setText("")
    # retranslateUi




import pytest
from mock import AsyncMock, ANY, MagicMock

from randovania.gui.dialog.racetime_browser_dialog import RacetimeBrowserDialog


@pytest.mark.parametrize("success", [False, True])
@pytest.mark.asyncio
async def test_attempt_join_success(skip_qtbot, mocker, success):
    # Setup
    permalink = MagicMock()

    def from_str(s: str):
        if s == "<permalink>":
            return permalink
        raise ValueError("Invalid permalink")

    raw_data = {
        "current_races": [
            {
                "name": "Race 1",
                "status": {
                    "value": "open",
                    "verbose_value": "Open",
                    "help_text": "Anyone may join this race"
                },
                "goal": {
                    "name": "Beat the game",
                    "custom": False
                },
                "info": "<permalink>",
                "entrants_count": 1,
                "entrants_count_inactive": 0,
                "opened_at": "2020-10-01T18:00:00.000Z"
            },
            {
                "name": "Race 2",
                "status": {
                    "value": "open",
                    "verbose_value": "Open",
                    "help_text": "Anyone may join this race"
                },
                "goal": {
                    "name": "Beat the game",
                    "custom": False
                },
                "info": "random_stuff",
                "entrants_count": 1,
                "entrants_count_inactive": 0,
                "opened_at": "2020-10-01T18:00:00.000Z"
            }
        ]
    }

    mock_warning = mocker.patch("randovania.gui.lib.async_dialog.warning", new_callable=AsyncMock)
    mock_from_str = mocker.patch("randovania.layout.permalink.Permalink.from_str", side_effect=from_str)
    mocker.patch("randovania.gui.dialog.racetime_browser_dialog._query_server", new_callable=AsyncMock,
                 return_value=raw_data)
    dialog = RacetimeBrowserDialog()

    # Run
    await dialog.refresh()
    dialog.table_widget.selectRow(0 if success else 1)
    await dialog.attempt_join()

    # Assert
    mock_from_str.assert_called_once()
    if success:
        mock_warning.assert_not_called()
        assert dialog.permalink is permalink
    else:
        mock_warning.assert_awaited_once_with(dialog, ANY, ANY)
        assert dialog.permalink is None



import dataclasses
from unittest.mock import MagicMock

import pytest

from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.echoes_configuration import LayoutSkyTempleKeyMode
from randovania.layout.elevators import LayoutElevators


@pytest.fixture(name="editor")
def _editor() -> PresetEditor:
    return PresetEditor(MagicMock())


_sample_layout_configurations = [
    {
        "sky_temple_keys": LayoutSkyTempleKeyMode.default(),
        "elevators": LayoutElevators.TWO_WAY_RANDOMIZED,
    }
]


@pytest.fixture(params=_sample_layout_configurations, name="initial_layout_configuration_params")
def _initial_layout_configuration_params(request) -> dict:
    return request.param


@pytest.mark.parametrize("skip_final_bosses", [False, True])
def test_edit_skip_final_bosses(editor: PresetEditor,
                                initial_layout_configuration_params: dict,
                                default_layout_configuration,
                                skip_final_bosses):
    # Setup
    editor._configuration = dataclasses.replace(default_layout_configuration,
                                                **initial_layout_configuration_params)
    editor._nested_autosave_level = 1

    # Run
    initial_layout_configuration_params["skip_final_bosses"] = skip_final_bosses
    editor.set_configuration_field("skip_final_bosses", skip_final_bosses)

    # Assert
    assert editor.configuration == dataclasses.replace(default_layout_configuration,
                                                       **initial_layout_configuration_params)



import json
from pathlib import Path

import pytest
from mock import MagicMock, AsyncMock, call

from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.gui.multiworld_client import MultiworldClient, Data


@pytest.fixture(name="client")
def _client(skip_qtbot):
    network_client = MagicMock()
    game_connection = MagicMock()
    game_connection.backend.lock_identifier = None
    return MultiworldClient(network_client, game_connection)


@pytest.mark.asyncio
async def test_start(client, tmpdir):
    game_connection = client.game_connection

    client.network_client.game_session_request_pickups = AsyncMock(return_value=[])
    client.network_client.session_self_update = AsyncMock()
    client.refresh_received_pickups = AsyncMock()
    client._received_messages = ["Foo"]
    client._received_pickups = ["Pickup"]

    # Run
    await client.start(Path(tmpdir).joinpath("missing_file.json"))

    # Assert
    client.refresh_received_pickups.assert_awaited_once_with()
    game_connection.set_location_collected_listener.assert_called_once_with(client.on_location_collected)
    client.network_client.GameUpdateNotification.connect.assert_called_once_with(client.on_network_game_updated)
    game_connection.set_permanent_pickups.assert_called_once_with(["Pickup"])


@pytest.mark.asyncio
async def test_stop(client):
    # Run
    await client.stop()

    # Assert
    client.game_connection.set_location_collected_listener.assert_called_once_with(None)
    client.network_client.GameUpdateNotification.disconnect.assert_called_once_with(client.on_network_game_updated)
    client.game_connection.set_permanent_pickups.assert_called_once_with([])


@pytest.mark.parametrize("exists", [False, True])
@pytest.mark.asyncio
async def test_on_location_collected(client, tmpdir, exists):
    client._data = Data(Path(tmpdir).joinpath("data.json"))
    client._data.collected_locations = {10, 15} if exists else {10}
    client.start_notify_collect_locations_task = MagicMock()

    # Run
    await client.on_location_collected(15)

    # Assert
    assert client._data.collected_locations == {10, 15}

    if exists:
        client.start_notify_collect_locations_task.assert_not_called()
    else:
        client.start_notify_collect_locations_task.assert_called_once_with()


@pytest.mark.asyncio
async def test_refresh_received_pickups(client):
    results = [
        ("Message A", b"bytesA"),
        ("Message B", b"bytesB"),
        ("Message C", b"bytesC"),
    ]

    client.network_client.game_session_request_pickups = AsyncMock(return_value=results)

    pickups = [MagicMock(), MagicMock(), MagicMock()]
    client._decode_pickup = MagicMock(side_effect=pickups)

    # Run
    await client.refresh_received_pickups()

    # Assert
    assert client._received_pickups == list(zip(["Message A", "Message B", "Message C"], pickups))
    client._decode_pickup.assert_has_calls([call(b"bytesA"), call(b"bytesB"), call(b"bytesC")])


@pytest.mark.asyncio
async def test_on_game_updated(client, tmpdir):
    client.refresh_received_pickups = AsyncMock()
    client._received_pickups = MagicMock()

    client._data = Data(Path(tmpdir).joinpath("data.json"))

    # Run
    await client.on_network_game_updated()

    # Assert
    client.game_connection.set_permanent_pickups.assert_called_once_with(client._received_pickups)


def test_decode_pickup(client, echoes_resource_database):
    data = b"\x80*42\x90'0\xb6\xb2\x80d \x00"
    expected_pickup = PickupEntry(
        name="The Name",
        model_index=0,
        item_category=ItemCategory.MOVEMENT,
        broad_category=ItemCategory.MOVEMENT,
        resources=(
            ConditionalResources(None, None, ()),
        ),
    )

    # from randovania.bitpacking import bitpacking
    # from randovania.network_common.pickup_serializer import BitPackPickupEntry
    # new_data = bitpacking.pack_value(BitPackPickupEntry(expected_pickup, echoes_resource_database))
    # assert new_data == data

    # Run
    pickup = client._decode_pickup(data)

    # Assert
    assert pickup == expected_pickup


@pytest.mark.asyncio
async def test_notify_collect_locations(client, tmpdir):
    data_path = Path(tmpdir).joinpath("data.json")
    network_client = client.network_client
    network_client.game_session_collect_locations = AsyncMock(side_effect=[
        RuntimeError("connection issue!"),
        None,
    ])

    data_path.write_text(json.dumps({
        "collected_locations": [10, 15],
        "uploaded_locations": [15],
        "latest_message_displayed": 0,
    }))
    client._data = Data(data_path)

    # Run
    await client._notify_collect_locations()

    # Assert
    network_client.game_session_collect_locations.assert_has_awaits([call((10,)), call((10,))])
    assert set(json.loads(data_path.read_text())["uploaded_locations"]) == {10, 15}


@pytest.mark.asyncio
async def test_lock_file_on_init(skip_qtbot, tmpdir):
    # Setup
    network_client = MagicMock()
    network_client.game_session_request_pickups = AsyncMock(return_value=[])
    network_client.session_self_update = AsyncMock()
    game_connection = MagicMock()
    game_connection.backend.lock_identifier = str(tmpdir.join("my-lock"))

    # Run
    client = MultiworldClient(network_client, game_connection)
    assert tmpdir.join("my-lock.pid").exists()

    await client.start(Path(tmpdir).joinpath("data.json"))
    await client.stop()
    assert not tmpdir.join("my-lock.pid").exists()



import re
import struct
from typing import Optional

import pytest
from mock import AsyncMock, MagicMock

from randovania.game_connection.connection_backend import ConnectionBackend, MemoryOperation, MemoryOperationException
from randovania.game_connection.connection_base import InventoryItem
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.pickup_entry import PickupEntry, ConditionalResources
from randovania.games.prime import dol_patcher


@pytest.fixture(name="backend")
def dolphin_backend():
    backend = ConnectionBackend()
    backend._perform_memory_operations = AsyncMock()
    return backend


def add_memory_op_result(backend, result):
    backend._perform_memory_operations.side_effect = lambda ops: {ops[0]: result}


@pytest.mark.asyncio
async def test_identify_game_ntsc(backend):
    # Setup
    add_memory_op_result(backend, b"!#$MetroidBuildInfo!#$Build v1.028 10/18/2004 10:44:32")

    # Run
    assert await backend._identify_game()

    # Assert
    assert backend.patches is dol_patcher.ALL_VERSIONS_PATCHES[0]


@pytest.mark.asyncio
async def test_identify_game_error(backend):
    # Setup
    backend._perform_memory_operations.side_effect = RuntimeError("not connected")

    # Run
    assert not await backend._identify_game()


@pytest.mark.asyncio
async def test_identify_game_already_known(backend):
    # Setup
    backend.patches = True
    backend._perform_memory_operations.side_effect = RuntimeError("not connected")

    # Run
    assert await backend._identify_game()


@pytest.mark.parametrize(["message_original", "message_encoded", "previous_size"], [
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00\x00', 0),
    ("Magoo2", b'\x00M\x00a\x00g\x00o\x00o\x002\x00\x00\x00\x00', 0),
    ("Magoo", b'\x00M\x00a\x00g\x00o\x00o\x00 \x00\x00\x00\x00', 10),
])
@pytest.mark.asyncio
async def test_write_string_to_game_buffer(backend, message_original, message_encoded, previous_size):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._last_message_size = previous_size

    # Run
    result = backend._write_string_to_game_buffer(message_original)

    # Assert
    assert result == MemoryOperation(backend.patches.string_display.message_receiver_string_ref,
                                     write_bytes=message_encoded)


@pytest.mark.asyncio
async def test_get_inventory_valid(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II", item.max_capacity, item.max_capacity)
        for op, item in zip(ops, backend.game.resource_database.item)
    }

    # Run
    inventory = await backend._get_inventory()

    # Assert
    assert inventory == {
        item: InventoryItem(item.max_capacity, item.max_capacity)
        for item in backend.game.resource_database.item
    }


@pytest.mark.asyncio
async def test_get_inventory_invalid_capacity(backend):
    # Setup
    custom_inventory = {5: InventoryItem(0, 50)}

    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II", *custom_inventory.get(item.index, InventoryItem(item.max_capacity, item.max_capacity)))
        for op, item in zip(ops, backend.game.resource_database.item)
    }

    # Run
    msg = "Received InventoryItem(amount=0, capacity=50) for Darkburst, which is an invalid state."
    with pytest.raises(MemoryOperationException, match=re.escape(msg)):
        await backend._get_inventory()


@pytest.mark.asyncio
async def test_get_inventory_invalid_amount(backend):
    # Setup
    custom_inventory = {5: InventoryItem(1, 0)}

    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._perform_memory_operations.side_effect = lambda ops: {
        op: struct.pack(">II", *custom_inventory.get(item.index, InventoryItem(item.max_capacity, item.max_capacity)))
        for op, item in zip(ops, backend.game.resource_database.item)
    }

    # Run
    msg = "Received InventoryItem(amount=1, capacity=0) for Darkburst, which is an invalid state."
    with pytest.raises(MemoryOperationException, match=re.escape(msg)):
        await backend._get_inventory()



@pytest.mark.asyncio
async def test_update_magic_item_nothing(backend):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._inventory = {
        backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)
    }
    backend._emit_location_collected = AsyncMock()
    backend._execute_remote_patches = AsyncMock()

    # Run
    await backend._update_magic_item()

    # Assert
    backend._emit_location_collected.assert_not_awaited()
    backend._execute_remote_patches.assert_not_awaited()


@pytest.mark.asyncio
async def test_update_magic_item_collected_location(backend, mocker):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.adjust_item_amount_and_capacity_patch")
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._inventory = {
        backend.game.resource_database.multiworld_magic_item: InventoryItem(10, 10)
    }
    backend._emit_location_collected = AsyncMock()
    backend._execute_remote_patches = AsyncMock()

    # Run
    await backend._update_magic_item()

    # Assert
    mock_item_patch.assert_called_once_with(backend.patches.powerup_functions,
                                            backend.game.resource_database.multiworld_magic_item.index,
                                            -10)
    backend._emit_location_collected.assert_awaited_once_with(9)
    backend._execute_remote_patches.assert_awaited_once_with([mock_item_patch.return_value], None)


@pytest.mark.parametrize("on_cooldown", [False, True])
@pytest.mark.asyncio
async def test_update_magic_item_give_pickup(backend, mocker, on_cooldown):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.increment_item_capacity_patch")
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._emit_location_collected = AsyncMock()
    backend._execute_remote_patches = AsyncMock()
    pickup_patches = MagicMock()
    backend._patches_for_pickup = AsyncMock(return_value=([pickup_patches, pickup_patches], "The Message"))

    backend._inventory = {backend.game.resource_database.multiworld_magic_item: InventoryItem(0, 0)}
    backend._permanent_pickups = [
        ("A", MagicMock()),
        ("B", MagicMock()),
    ]
    backend.message_cooldown = 1 if on_cooldown else 0

    # Run
    await backend._update_magic_item()

    # Assert
    backend._emit_location_collected.assert_not_awaited()
    if on_cooldown:
        backend._execute_remote_patches.assert_not_awaited()
    else:
        mock_item_patch.assert_called_once_with(backend.patches.powerup_functions,
                                                backend.game.resource_database.multiworld_magic_item.index)
        backend._patches_for_pickup.assert_awaited_once_with(*backend._permanent_pickups[0])
        backend._execute_remote_patches.assert_awaited_once_with(
            [pickup_patches, pickup_patches, mock_item_patch.return_value], "The Message")


@pytest.mark.asyncio
async def test_patches_for_pickup(backend, mocker):
    # Setup
    mock_item_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.adjust_item_amount_and_capacity_patch")
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]

    db = backend.game.resource_database
    pickup = PickupEntry("Pickup", 0, ItemCategory.MISSILE, ItemCategory.MISSILE, (
        ConditionalResources(None, None,
                             ((db.energy_tank, db.energy_tank.max_capacity),
                              (db.item_percentage, 1)),
                             ),
    ))
    backend._inventory = {
        db.multiworld_magic_item: InventoryItem(0, 0),
        db.energy_tank: InventoryItem(1, 1),
    }

    # Run
    patches, message = await backend._patches_for_pickup("Someone", pickup)

    # Assert
    mock_item_patch.assert_called_once_with(backend.patches.powerup_functions, db.energy_tank.index,
                                            db.energy_tank.max_capacity)
    assert patches == [mock_item_patch.return_value]
    assert message == "Received Pickup from Someone."


@pytest.mark.parametrize("has_message", [False, True])
@pytest.mark.asyncio
async def test_execute_remote_patches(backend, mocker, has_message):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    backend._write_string_to_game_buffer = MagicMock()

    patch_address, patch_bytes = MagicMock(), MagicMock()
    mock_remote_execute: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.create_remote_execution_body",
        return_value=(patch_address, patch_bytes)
    )
    mock_hud_patch: MagicMock = mocker.patch(
        "randovania.games.prime.all_prime_dol_patches.call_display_hud_patch",
        return_value=([7, 8])
    )

    patches = [[1, 2, 3], [4, 5, 6]]
    instructions = [1, 2, 3, 4, 5, 6]
    message = None
    message_ops = []

    if has_message:
        message = "A message to show!"
        instructions.extend([7, 8])
        message_ops.append(backend._write_string_to_game_buffer.return_value)

    memory_operations = [
        *message_ops,
        MemoryOperation(patch_address, write_bytes=patch_bytes),
        MemoryOperation(backend.patches.cstate_manager_global + 0x2, write_bytes=b"\x01"),
    ]

    # Run
    await backend._execute_remote_patches(patches, message)

    # Assert
    mock_remote_execute.assert_called_once_with(backend.patches.string_display, instructions)
    backend._perform_memory_operations.assert_awaited_once_with(memory_operations)
    if has_message:
        mock_hud_patch.assert_called_once_with(backend.patches.string_display)
        backend._write_string_to_game_buffer.assert_called_once_with(message)
        assert backend.message_cooldown == 4
    else:
        backend._write_string_to_game_buffer.assert_not_called()
        assert backend.message_cooldown == 0


@pytest.mark.parametrize("correct_vtable", [False, True])
@pytest.mark.parametrize("has_pending_op", [False, True])
@pytest.mark.parametrize("has_world", [False, True])
@pytest.mark.asyncio
async def test_fetch_game_status(backend, has_world, has_pending_op, correct_vtable):
    # Setup
    backend.patches = dol_patcher.ALL_VERSIONS_PATCHES[0]
    world = backend.game.world_list.worlds[0]

    backend._perform_memory_operations.side_effect = lambda ops: {
        ops[0]: world.world_asset_id.to_bytes(4, "big") if has_world else b"DEAD",
        ops[1]: b"\x01" if has_pending_op else b"\x00",
        ops[2]: backend.patches.cplayer_vtable.to_bytes(4, "big") if correct_vtable else b"CAFE",
    }

    # Run
    await backend._fetch_game_status()

    # Assert
    if has_world and correct_vtable:
        assert backend._world is world
    else:
        assert backend._world is None
    assert backend._has_pending_op is has_pending_op


@pytest.mark.parametrize("failure_at", [None, 1, 2])
@pytest.mark.parametrize("depth", [0, 1, 2])
@pytest.mark.asyncio
async def test_interact_with_game(backend, depth: int, failure_at: Optional[int]):
    # Setup
    backend._fetch_game_status = AsyncMock()
    backend._world = MagicMock() if depth > 0 else None
    backend._get_inventory = AsyncMock()
    if failure_at == 1:
        backend._get_inventory.side_effect = MemoryOperationException("error at _get_inventory")
    backend._has_pending_op = depth <= 1
    backend._update_magic_item = AsyncMock()
    if failure_at == 2:
        backend._update_magic_item.side_effect = MemoryOperationException("error at _check_for_collected_index")
    backend.message_cooldown = 2

    expected_depth = min(depth, failure_at) if failure_at is not None else depth

    # Run
    await backend._interact_with_game(1)

    # Assert
    assert backend.message_cooldown == (2 if expected_depth < 2 else 1)
    backend._fetch_game_status.assert_awaited_once_with()

    if expected_depth > 0:
        backend._get_inventory.assert_awaited_once_with()
    else:
        backend._get_inventory.assert_not_awaited()

    if expected_depth > 1:
        backend._update_magic_item.assert_awaited_once_with()
    else:
        backend._update_magic_item.assert_not_awaited()

    if 0 < depth < (failure_at or 999):
        assert backend._world is not None
    else:
        assert backend._world is None



from typing import Optional, List, Union

from PySide2 import QtWidgets
from PySide2.QtCore import Qt
from PySide2.QtGui import QIntValidator
from PySide2.QtWidgets import QDialog
from PySide2.QtWidgets import QPushButton, QWidget, QGroupBox, QVBoxLayout, QHBoxLayout, QComboBox, \
    QLineEdit

from randovania.game_description.requirements import ResourceRequirement, Requirement, \
    RequirementOr, RequirementAnd, RequirementTemplate
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.resource_type import ResourceType
from randovania.gui.generated.connections_editor_ui import Ui_ConnectionEditor
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.interface_common.enum_lib import iterate_enum


def _create_resource_name_combo(resource_database: ResourceDatabase,
                                resource_type: ResourceType,
                                current_resource: Optional[ResourceInfo],
                                parent: QWidget,
                                ) -> QComboBox:
    """

    :param resource_database:
    :param current_resource:
    :param parent:
    :return:
    """

    resource_name_combo = QComboBox(parent)

    for resource in sorted(resource_database.get_by_type(resource_type), key=lambda x: x.long_name):
        resource_name_combo.addItem(resource.long_name, resource)
        if resource is current_resource:
            resource_name_combo.setCurrentIndex(resource_name_combo.count() - 1)

    return resource_name_combo


def _create_resource_type_combo(current_resource_type: ResourceType, parent: QWidget,
                                resource_database: ResourceDatabase) -> QComboBox:
    """

    :param current_resource_type:
    :param parent:
    :return:
    """
    resource_type_combo = QComboBox(parent)

    for resource_type in iterate_enum(ResourceType):
        try:
            count_elements = len(resource_database.get_by_type(resource_type))
        except ValueError:
            count_elements = 0

        if count_elements == 0:
            continue

        resource_type_combo.addItem(resource_type.name, resource_type)
        if resource_type is current_resource_type:
            resource_type_combo.setCurrentIndex(resource_type_combo.count() - 1)

    return resource_type_combo


def _create_default_resource_requirement(resource_database: ResourceDatabase) -> ResourceRequirement:
    return ResourceRequirement(
        resource_database.get_by_type(ResourceType.ITEM)[0],
        1, False
    )


def _create_default_template_requirement(resource_database: ResourceDatabase) -> RequirementTemplate:
    template_name = None
    for template_name in resource_database.requirement_template.keys():
        break
    if template_name is None:
        raise RuntimeError("No templates?!")
    return RequirementTemplate(resource_database, template_name)


class ResourceRequirementEditor:
    def __init__(self,
                 parent: QWidget, layout: QHBoxLayout,
                 resource_database: ResourceDatabase, item: ResourceRequirement,
                 ):
        self.parent = parent
        self.layout = layout
        self.resource_database = resource_database

        self.resource_type_combo = _create_resource_type_combo(item.resource.resource_type, parent, resource_database)
        self.resource_type_combo.setMinimumWidth(75)
        self.resource_type_combo.setMaximumWidth(75)

        self.resource_name_combo = _create_resource_name_combo(self.resource_database,
                                                               item.resource.resource_type,
                                                               item.resource,
                                                               self.parent)

        self.negate_combo = QComboBox(parent)
        self.negate_combo.addItem("≥", False)
        self.negate_combo.addItem("<", True)
        self.negate_combo.setCurrentIndex(int(item.negate))
        self.negate_combo.setMinimumWidth(40)
        self.negate_combo.setMaximumWidth(40)

        self.amount_edit = QLineEdit(parent)
        self.amount_edit.setValidator(QIntValidator(1, 10000))
        self.amount_edit.setText(str(item.amount))
        self.amount_edit.setMinimumWidth(45)
        self.amount_edit.setMaximumWidth(45)

        self.layout.addWidget(self.resource_type_combo)
        self.layout.addWidget(self.resource_name_combo)
        self.layout.addWidget(self.negate_combo)
        self.layout.addWidget(self.amount_edit)

        self.resource_type_combo.currentIndexChanged.connect(self._update_type)

    def _update_type(self):
        old_combo = self.resource_name_combo
        self.resource_name_combo = _create_resource_name_combo(self.resource_database,
                                                               self.resource_type_combo.currentData(),
                                                               None,
                                                               self.parent)

        self.layout.replaceWidget(old_combo, self.resource_name_combo)
        old_combo.deleteLater()

    def deleteLater(self):
        self.resource_type_combo.deleteLater()
        self.resource_name_combo.deleteLater()
        self.negate_combo.deleteLater()
        self.amount_edit.deleteLater()

    @property
    def current_requirement(self) -> ResourceRequirement:
        return ResourceRequirement(
            self.resource_name_combo.currentData(),
            int(self.amount_edit.text()),
            self.negate_combo.currentData()
        )


class ArrayRequirementEditor:
    def __init__(self, parent: QWidget, parent_layout: QVBoxLayout, resource_database: ResourceDatabase,
                 requirement: Union[RequirementOr, RequirementAnd]):
        self._editors = []
        self.resource_database = resource_database
        self._array_type = type(requirement)

        self.group_box = QGroupBox(parent)
        self.group_box.setStyleSheet("QGroupBox { margin-top: 2px; }")
        parent_layout.addWidget(self.group_box)
        self.item_layout = QVBoxLayout(self.group_box)
        self.item_layout.setContentsMargins(8, 2, 2, 6)
        self.item_layout.setAlignment(Qt.AlignTop)

        self.new_item_button = QPushButton(self.group_box)
        self.new_item_button.setMaximumWidth(75)
        self.new_item_button.setText("New Row")
        self.new_item_button.clicked.connect(self.new_item)

        for item in requirement.items:
            self._create_item(item)

        self.item_layout.addWidget(self.new_item_button)

    def _create_item(self, item: Requirement):
        def on_remove():
            self._editors.remove(nested_editor)
            nested_editor.deleteLater()

        nested_editor = RequirementEditor(self.group_box, self.item_layout, self.resource_database, on_remove=on_remove)
        nested_editor.create_specialized_editor(item)
        self._editors.append(nested_editor)

    def new_item(self):
        self._create_item(_create_default_resource_requirement(self.resource_database))

        self.item_layout.removeWidget(self.new_item_button)
        self.item_layout.addWidget(self.new_item_button)

    def deleteLater(self):
        self.group_box.deleteLater()
        for editor in self._editors:
            editor.deleteLater()
        self.new_item_button.deleteLater()

    @property
    def current_requirement(self) -> Union[RequirementOr, RequirementAnd]:
        return self._array_type([
            editor.current_requirement
            for editor in self._editors
        ])


class TemplateRequirementEditor:
    def __init__(self,
                 parent: QWidget, layout: QHBoxLayout,
                 resource_database: ResourceDatabase, item: RequirementTemplate,
                 ):
        self.parent = parent
        self.layout = layout
        self.resource_database = resource_database

        template_name_combo = QComboBox(parent)

        for template_name in sorted(resource_database.requirement_template.keys()):
            template_name_combo.addItem(template_name)
            if template_name == item.template_name:
                template_name_combo.setCurrentIndex(template_name_combo.count() - 1)

        self.template_name_combo = template_name_combo
        self.layout.addWidget(self.template_name_combo)

    def deleteLater(self):
        self.template_name_combo.deleteLater()

    @property
    def current_requirement(self) -> RequirementTemplate:
        return RequirementTemplate(
            self.resource_database,
            self.template_name_combo.currentText(),
        )


class RequirementEditor:
    _editor: Union[None, ResourceRequirementEditor, ArrayRequirementEditor, TemplateRequirementEditor]

    def __init__(self,
                 parent: QWidget,
                 parent_layout: QVBoxLayout,
                 resource_database: ResourceDatabase,
                 *, on_remove=None):

        self.parent = parent
        self.parent_layout = parent_layout
        self.resource_database = resource_database
        self._editor = None
        self._last_resource = None
        self._last_items = ()

        self.line_layout = QHBoxLayout()
        self.line_layout.setAlignment(Qt.AlignLeft)
        self.parent_layout.addLayout(self.line_layout)

        if on_remove is not None:
            self.remove_button = QtWidgets.QToolButton(parent)
            self.remove_button.setText("X")
            self.remove_button.setMaximumWidth(20)
            self.remove_button.clicked.connect(on_remove)
            self.line_layout.addWidget(self.remove_button)
        else:
            self.remove_button = None

        self.requirement_type_combo = QComboBox(parent)
        self.requirement_type_combo.addItem("Resource", ResourceRequirement)
        self.requirement_type_combo.addItem("Or", RequirementOr)
        self.requirement_type_combo.addItem("And", RequirementAnd)
        if resource_database.requirement_template:
            self.requirement_type_combo.addItem("Template", RequirementTemplate)
        self.requirement_type_combo.setMaximumWidth(75)
        self.requirement_type_combo.activated.connect(self._on_change_requirement_type)
        self.line_layout.addWidget(self.requirement_type_combo)

    def create_specialized_editor(self, requirement: Requirement):
        self.requirement_type_combo.setCurrentIndex(self.requirement_type_combo.findData(type(requirement)))

        if isinstance(requirement, ResourceRequirement):
            self._editor = ResourceRequirementEditor(self.parent, self.line_layout, self.resource_database, requirement)

        elif isinstance(requirement, (RequirementOr, RequirementAnd)):
            self._editor = ArrayRequirementEditor(self.parent, self.parent_layout, self.resource_database, requirement)

        elif isinstance(requirement, RequirementTemplate):
            self._editor = TemplateRequirementEditor(self.parent, self.line_layout, self.resource_database, requirement)

        else:
            raise RuntimeError(f"Unknown requirement type: {type(requirement)} - {requirement}")

    def _on_change_requirement_type(self):
        current_requirement = self.current_requirement
        self._editor.deleteLater()

        if isinstance(current_requirement, ResourceRequirement):
            self._last_resource = current_requirement

        elif isinstance(current_requirement, (RequirementOr, RequirementAnd)):
            self._last_items = current_requirement.items

        elif isinstance(current_requirement, RequirementTemplate):
            pass

        else:
            raise RuntimeError(f"Unknown requirement type: {type(current_requirement)} - {current_requirement}")

        new_class = self.requirement_type_combo.currentData()
        if new_class == ResourceRequirement:
            if self._last_resource is None:
                new_requirement = _create_default_resource_requirement(self.resource_database)
            else:
                new_requirement = self._last_resource
        elif new_class == RequirementTemplate:
            new_requirement = _create_default_template_requirement(self.resource_database)
        else:
            new_requirement = new_class(self._last_items)

        self.create_specialized_editor(new_requirement)

    def deleteLater(self):
        if self.remove_button is not None:
            self.remove_button.deleteLater()

        self.requirement_type_combo.deleteLater()

        if self._editor is not None:
            self._editor.deleteLater()

    @property
    def current_requirement(self) -> Requirement:
        return self._editor.current_requirement


class ConnectionsEditor(QDialog, Ui_ConnectionEditor):
    parent: QWidget
    resource_database: ResourceDatabase
    _elements: List[QWidget]

    def __init__(self, parent: QWidget, resource_database: ResourceDatabase, requirement: Requirement):
        super().__init__(parent)
        self.setupUi(self)
        set_default_window_icon(self)

        self.parent = parent
        self.resource_database = resource_database

        self.contents_layout.setAlignment(Qt.AlignTop)

        self._root_editor = RequirementEditor(self, self.contents_layout, resource_database)
        self._root_editor.create_specialized_editor(requirement)

    def deleteLater(self):
        self._root_editor.deleteLater()

    def build_requirement(self) -> Requirement:
        return self._root_editor.current_requirement

    @property
    def final_requirement(self) -> Optional[Requirement]:
        result = self.build_requirement()
        if result == Requirement.impossible():
            return None
        return result



import asyncio
import dataclasses
import functools
import json
import logging
import random
from typing import List, Optional

from PySide2 import QtWidgets, QtGui
from PySide2.QtCore import Qt, QTimer
from PySide2.QtWidgets import QMessageBox
from asyncqt import asyncSlot, asyncClose

from randovania.game_connection.game_connection import GameConnection
from randovania.game_description import data_reader
from randovania.games.game import RandovaniaGame
from randovania.generator import base_patches_factory
from randovania.gui.dialog.echoes_user_preferences_dialog import EchoesUserPreferencesDialog
from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.gui.dialog.permalink_dialog import PermalinkDialog
from randovania.gui.generated.game_session_ui import Ui_GameSessionWindow
from randovania.gui.lib import common_qt_lib, preset_describer, async_dialog
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.game_connection_setup import GameConnectionSetup
from randovania.gui.lib.generation_failure_handling import GenerationFailureHandler
from randovania.gui.lib.qt_network_client import handle_network_errors, QtNetworkClient
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.multiworld_client import MultiworldClient, BackendInUse
from randovania.gui.preset_settings.logic_settings_window import LogicSettingsWindow
from randovania.interface_common import simplified_patcher, status_update_lib
from randovania.interface_common.options import Options, InfoAlert
from randovania.interface_common.preset_editor import PresetEditor
from randovania.interface_common.preset_manager import PresetManager
from randovania.interface_common.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.layout.preset_migration import VersionedPreset
from randovania.network_client.game_session import GameSessionEntry, PlayerSessionEntry
from randovania.network_client.network_client import ConnectionState
from randovania.network_common.admin_actions import SessionAdminUserAction, SessionAdminGlobalAction
from randovania.network_common.session_state import GameSessionState
from randovania.resolver.exceptions import GenerationFailure

logger = logging.getLogger(__name__)


@dataclasses.dataclass()
class PlayerWidget:
    name: QtWidgets.QLabel
    connection_state: Optional[QtWidgets.QLabel]
    tool: QtWidgets.QToolButton
    kick: QtWidgets.QAction
    promote: QtWidgets.QAction
    open_tracker: QtWidgets.QAction
    abandon: QtWidgets.QAction
    move_up: QtWidgets.QAction
    move_down: QtWidgets.QAction
    switch_observer_action: QtWidgets.QAction
    player: Optional[PlayerSessionEntry] = None

    @property
    def widgets(self) -> List[QtWidgets.QWidget]:
        result = [self.name, self.tool]
        if self.connection_state is not None:
            result.append(self.connection_state)
        return result

    def delete_widgets(self):
        for widget in self.widgets:
            widget.deleteLater()

    def set_visible(self, value: bool):
        for widget in self.widgets:
            widget.setVisible(value)

    def set_player(self, player: Optional[PlayerSessionEntry]):
        self.player = player
        if player is not None:
            self.set_visible(True)
            for action in (self.move_up, self.move_down, self.abandon, self.open_tracker):
                action.setVisible(not player.is_observer)
        else:
            self.set_visible(False)

    def update_player_widget(self,
                             game_session: GameSessionEntry,
                             self_player: PlayerSessionEntry,
                             ):

        player = self.player
        if player is None:
            return

        player_name = player.name
        if player.id == self_player.id:
            player_name += " (You)"
            admin_or_you = True
        else:
            admin_or_you = self_player.admin

        player_is_admin = game_session.players[self.player.id].admin
        if player_is_admin:
            promote_text = "Demote from Admin"
            num_required = 2
            player_name = "[Admin] " + player_name
        else:
            promote_text = "Promote to Admin"
            num_required = 0

        self.name.setText(player_name)
        if self.connection_state is not None:
            self.connection_state.setText(player.connection_state)
        self.kick.setEnabled(self_player.admin and player.id != self_player.id)
        self.promote.setText(promote_text)
        self.promote.setEnabled((self_player.admin or game_session.num_admins == 0)
                                and game_session.num_admins >= num_required)
        if not player.is_observer:
            self.move_up.setEnabled(admin_or_you and player.row > 0)
            self.move_down.setEnabled(admin_or_you and player.row + 1 < game_session.num_rows)
        self.abandon.setEnabled(admin_or_you)
        self.switch_observer_action.setText("Include in session" if player.is_observer else "Move to observers")
        self.switch_observer_action.setEnabled(admin_or_you)


@dataclasses.dataclass()
class RowWidget:
    name: QtWidgets.QLabel
    tool: QtWidgets.QToolButton
    view_summary: QtWidgets.QAction
    customize: QtWidgets.QAction
    import_menu: QtWidgets.QMenu
    import_actions: List[QtWidgets.QAction]
    save_copy: QtWidgets.QAction
    delete: QtWidgets.QAction

    @property
    def widgets(self):
        return [self.name, self.tool]

    def delete_widgets(self):
        for widget in self.widgets:
            widget.deleteLater()

    def set_is_admin(self, is_admin: bool, is_your_row: bool, can_customize: bool):
        for widget in (self.customize, self.import_menu):
            widget.setEnabled(can_customize and (is_admin or is_your_row))
        self.delete.setEnabled(can_customize and is_admin)

    def set_preset(self, preset: Preset):
        self.name.setText(preset.name)


_PRESET_COLUMNS = 3


class GameSessionWindow(QtWidgets.QMainWindow, Ui_GameSessionWindow, BackgroundTaskMixin):
    team_players: List[PlayerWidget]
    observers: List[PlayerWidget]
    rows: List[RowWidget]
    _game_session: GameSessionEntry
    has_closed = False
    _logic_settings_window: Optional[LogicSettingsWindow] = None
    _window_manager: WindowManager
    _generating_game: bool = False
    _expecting_kick = False
    _already_kicked = False

    def __init__(self, network_client: QtNetworkClient, game_connection: GameConnection,
                 preset_manager: PresetManager, window_manager: WindowManager, options: Options):
        super().__init__()
        self.setupUi(self)
        common_qt_lib.set_default_window_icon(self)

        self.network_client = network_client
        self.game_connection = game_connection
        self.multiworld_client = MultiworldClient(self.network_client, self.game_connection)
        self.failure_handler = GenerationFailureHandler(self)

        self._preset_manager = preset_manager
        self._window_manager = window_manager
        self._options = options
        self._update_status_lock = asyncio.Lock()

        # Advanced Options
        self.advanced_options_menu = QtWidgets.QMenu(self.advanced_options_tool)
        self.rename_session_action = QtWidgets.QAction("Change title", self.advanced_options_menu)
        self.change_password_action = QtWidgets.QAction("Change password", self.advanced_options_menu)
        self.delete_session_action = QtWidgets.QAction("Delete session", self.advanced_options_menu)

        self.advanced_options_menu.addAction(self.rename_session_action)
        self.advanced_options_menu.addAction(self.change_password_action)
        self.advanced_options_menu.addAction(self.delete_session_action)
        self.advanced_options_tool.setMenu(self.advanced_options_menu)

        self.rename_session_action.triggered.connect(self.rename_session)
        self.change_password_action.triggered.connect(self.change_password)
        self.delete_session_action.triggered.connect(self.delete_session)

        # Save ISO Button
        self.save_iso_menu = QtWidgets.QMenu(self.save_iso_button)
        self.copy_permalink_action = QtWidgets.QAction("Copy Permalink", self.save_iso_menu)

        self.save_iso_menu.addAction(self.copy_permalink_action)
        self.save_iso_button.setMenu(self.save_iso_menu)

        self.copy_permalink_action.triggered.connect(self.copy_permalink)

        # Background process Button
        self.background_process_menu = QtWidgets.QMenu(self.background_process_button)
        self.generate_game_with_spoiler_action = QtWidgets.QAction("Generate game", self.background_process_menu)
        self.generate_game_without_spoiler_action = QtWidgets.QAction("Generate without spoiler",
                                                                      self.background_process_menu)
        self.import_permalink_action = QtWidgets.QAction("Import permalink", self.background_process_menu)
        self.background_process_menu.addAction(self.generate_game_with_spoiler_action)
        self.background_process_menu.addAction(self.generate_game_without_spoiler_action)
        self.background_process_menu.addAction(self.import_permalink_action)
        self.background_process_button.setMenu(self.background_process_menu)

        self.generate_game_with_spoiler_action.triggered.connect(self.generate_game_with_spoiler)
        self.generate_game_without_spoiler_action.triggered.connect(self.generate_game_without_spoiler)
        self.import_permalink_action.triggered.connect(self.import_permalink)
        self.background_process_button.clicked.connect(self.background_process_button_clicked)

        # Signals
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.customize_user_preferences_button.clicked.connect(self._open_user_preferences_dialog)
        self.session_status_tool.clicked.connect(self._session_status_button_clicked)
        self.save_iso_button.clicked.connect(self.save_iso)
        self.view_game_details_button.clicked.connect(self.view_game_details)

        # Game Connection
        self.game_connection_setup = GameConnectionSetup(self, self.game_connection_tool, self.game_connection_label,
                                                         self.game_connection, self._options)

        # Server Status
        self.server_connection_button.clicked.connect(self._connect_to_server)

        # Session status
        self.session_status_menu = QtWidgets.QMenu(self.session_status_tool)
        self.start_session_action = QtWidgets.QAction("Start session", self.session_status_menu)
        self.start_session_action.triggered.connect(self.start_session)

        self.finish_session_action = QtWidgets.QAction("Finish session", self.session_status_menu)
        self.finish_session_action.triggered.connect(self.finish_session)

        self.reset_session_action = QtWidgets.QAction("Reset session", self.session_status_menu)
        self.reset_session_action.triggered.connect(self.reset_session)

        self.session_status_menu.addAction(self.start_session_action)
        self.session_status_menu.addAction(self.finish_session_action)
        self.session_status_menu.addAction(self.reset_session_action)
        self.session_status_tool.setMenu(self.session_status_menu)

        self.players_layout.addWidget(self.players_vertical_line, 0, 2, -1, 1)
        self.new_row_button.clicked.connect(functools.partial(
            self._admin_global_action_slot, SessionAdminGlobalAction.CREATE_ROW,
            self._preset_manager.default_preset.as_json))

        self.rows = []
        self.observers = []
        self.team_players = []

        self.main_layout.insertWidget(1, self.players_box)

        spacer_item = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.main_layout.insertSpacerItem(2, spacer_item)

        self.network_client.GameSessionUpdated.connect(self.on_game_session_updated)
        self.network_client.ConnectionStateUpdated.connect(self.on_server_connection_state_updated)
        self.game_connection.Updated.connect(self.on_game_connection_updated)

    @classmethod
    async def create_and_update(cls, network_client: QtNetworkClient, game_connection: GameConnection,
                                preset_manager: PresetManager, window_manager: WindowManager, options: Options,
                                ) -> Optional["GameSessionWindow"]:

        try:
            window = cls(network_client, game_connection, preset_manager, window_manager, options)
            await window.on_game_session_updated(network_client.current_game_session)
            await window.on_game_connection_updated()
            window.on_server_connection_state_updated(network_client.connection_state)
            return window

        except BackendInUse as e:
            await async_dialog.warning(
                window_manager, "Existing Connection",
                "Another instance of Randovania is already connected to an online session.\n"
                "Please close it before continuing.\n\n"
                f"If this error continues, please delete the following file:\n{e.lock_file}")

    @asyncClose
    async def closeEvent(self, event: QtGui.QCloseEvent):
        return await self._on_close_event(event)

    async def _on_close_event(self, event: QtGui.QCloseEvent):
        user_response = QMessageBox.No

        is_kicked = self.network_client.current_user.id not in self._game_session.players
        if not is_kicked:
            user_response = await async_dialog.warning(
                self,
                "Leaving Game Session",
                ("Do you want to also leave the session?\n\n"
                 "Yes: Leave permanently, freeing a spot for others.\n"
                 "No: Close the window, but stay in the session. You can rejoin later.\n"
                 "Cancel: Do nothing\n"),
                QMessageBox.Yes | QMessageBox.No | QMessageBox.Cancel,
                QMessageBox.No
            )

        if user_response == QMessageBox.Cancel:
            event.ignore()
            return

        await self.multiworld_client.stop()
        try:
            self.network_client.GameSessionUpdated.disconnect(self.on_game_session_updated)
        except Exception as e:
            logging.exception(f"Unable to disconnect: {e}")
        try:
            self.network_client.ConnectionStateUpdated.disconnect(self.on_server_connection_state_updated)
        except Exception as e:
            logging.exception(f"Unable to disconnect: {e}")
        try:
            self.game_connection.Updated.disconnect(self.on_game_connection_updated)
        except Exception as e:
            logging.exception(f"Unable to disconnect: {e}")

        try:
            if user_response == QMessageBox.Yes or (not is_kicked and
                                                    not self.network_client.connection_state.is_disconnected):
                await self.network_client.leave_game_session(user_response == QMessageBox.Yes)
        finally:
            super().closeEvent(event)
        self.has_closed = True

    # Row Functions

    def add_row(self):
        preset_name = QtWidgets.QLabel(self.players_box)
        preset_name.setText("")
        preset_name.setMaximumWidth(180)
        self.players_layout.addWidget(preset_name, len(self.rows) + 2, 0)

        tool_button = QtWidgets.QToolButton(self.players_box)
        tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        tool_button.setText("...")
        self.players_layout.addWidget(tool_button, len(self.rows) + 2, 1)

        tool_button_menu = QtWidgets.QMenu(tool_button)
        tool_button.setMenu(tool_button_menu)
        import_menu = QtWidgets.QMenu(tool_button_menu)

        row = RowWidget(
            name=preset_name,
            tool=tool_button,
            view_summary=QtWidgets.QAction(tool_button_menu),
            customize=QtWidgets.QAction(tool_button_menu),
            import_menu=import_menu,
            import_actions=[],
            save_copy=QtWidgets.QAction(tool_button_menu),
            delete=QtWidgets.QAction(tool_button_menu),
        )
        self.rows.append(row)

        row.view_summary.setText("View preset summary")
        row.view_summary.triggered.connect(functools.partial(self._row_show_preset_summary, row))
        tool_button_menu.addAction(row.view_summary)

        row.customize.setText("Customize preset")
        row.customize.triggered.connect(functools.partial(self._row_customize_preset, row))
        tool_button_menu.addAction(row.customize)

        row.import_menu.setTitle("Import preset")
        tool_button_menu.addMenu(row.import_menu)
        self._create_actions_for_import_menu(row)

        row.save_copy.setText("Save copy of preset")
        row.save_copy.triggered.connect(functools.partial(self._row_save_preset, row))
        tool_button_menu.addAction(row.save_copy)

        row.delete.setText("Delete session row")
        row.delete.triggered.connect(functools.partial(self._row_delete, row))
        tool_button_menu.addAction(row.delete)

        self.append_new_player_widget(False)

    def pop_row(self):
        self.rows.pop().delete_widgets()
        while len(self.team_players) > len(self.rows):
            self.team_players.pop().delete_widgets()

    def _create_actions_for_import_menu(self, row: RowWidget):
        for included_preset in self._preset_manager.all_presets:
            if included_preset.game != RandovaniaGame.PRIME2:
                continue
            action = QtWidgets.QAction(row.import_menu)
            action.setText(included_preset.name)
            action.triggered.connect(functools.partial(self._row_import_preset, row, included_preset))
            row.import_actions.append(action)
            row.import_menu.addAction(action)

    def refresh_row_import_preset_actions(self):
        for row in self.rows:
            row.import_menu.clear()
            self._create_actions_for_import_menu(row)

    @asyncSlot()
    async def _row_show_preset_summary(self, row: RowWidget):
        row_index = self.rows.index(row)
        preset = self._game_session.presets[row_index].get_preset()
        description = preset_describer.merge_categories(preset_describer.describe(preset))

        message_box = QtWidgets.QMessageBox(self)
        message_box.setWindowTitle(preset.name)
        message_box.setText(description)
        message_box.setTextInteractionFlags(Qt.TextSelectableByMouse)
        await async_dialog.execute_dialog(message_box)

    @asyncSlot()
    @handle_network_errors
    async def _row_customize_preset(self, row: RowWidget):
        if self._logic_settings_window is not None:
            if self._logic_settings_window._game_session_row == row:
                self._logic_settings_window.raise_()
                self._logic_settings_window.activateWindow()
            else:
                # show warning that a dialog is already in progress?
                await async_dialog.warning(self, "Customize in progress",
                                           "A window for customizing a preset is already open. "
                                           "Please close it before continuing.",
                                           QtWidgets.QMessageBox.Ok, QtWidgets.QMessageBox.Ok)
            return

        row_index = self.rows.index(row)
        editor = PresetEditor(self._game_session.presets[row_index].get_preset())
        self._logic_settings_window = LogicSettingsWindow(None, editor)
        self._logic_settings_window._game_session_row = row

        self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())
        editor.on_changed = lambda: self._logic_settings_window.on_preset_changed(editor.create_custom_preset_with())

        result = await async_dialog.execute_dialog(self._logic_settings_window)
        self._logic_settings_window = None

        if result == QtWidgets.QDialog.Accepted:
            new_preset = VersionedPreset.with_preset(editor.create_custom_preset_with())
            if self._preset_manager.add_new_preset(new_preset):
                self.refresh_row_import_preset_actions()

            await self._admin_global_action(SessionAdminGlobalAction.CHANGE_ROW, (row_index, new_preset.as_json))

    @asyncSlot()
    @handle_network_errors
    async def _row_import_preset(self, row: RowWidget, preset: Preset):
        row_index = self.rows.index(row)
        await self._admin_global_action(SessionAdminGlobalAction.CHANGE_ROW, (row_index, preset.as_json))

    def _row_save_preset(self, row: RowWidget):
        row_index = self.rows.index(row)
        preset = self._game_session.presets[row_index]

        existing_preset = self._preset_manager.preset_for_name(preset.name)
        if existing_preset is not None:
            if existing_preset == preset:
                return

            user_response = QMessageBox.warning(
                self,
                "Preset name conflict",
                "A preset named '{}' already exists. Do you want to overwrite it?".format(preset.name),
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No
            )
            if user_response == QMessageBox.No:
                return

        if self._preset_manager.add_new_preset(preset):
            self.refresh_row_import_preset_actions()

    @asyncSlot()
    @handle_network_errors
    async def _row_delete(self, row: RowWidget):
        row_index = self.rows.index(row)
        if row != self.rows[-1]:
            await async_dialog.message_box(self, QMessageBox.Critical,
                                           "Unable to delete row", "Can only delete the last row.")
        elif len(self.rows) <= 1:
            await async_dialog.message_box(self, QMessageBox.Critical,
                                           "Unable to delete row", "At least one row must remain.")
        else:
            await self._admin_global_action(SessionAdminGlobalAction.DELETE_ROW, row_index)

    # Team Functions
    def append_new_player_widget(self, is_observer: bool):
        if is_observer:
            parent_layout = self.observer_layout
            observer_index = len(self.observers)
            team_id = observer_index % 3
            row_id = observer_index // 3
        else:
            parent_layout = self.players_layout
            team_id = 0
            row_id = len(self.team_players)

        player_label = QtWidgets.QLabel(self.players_box)
        player_label.setText("")
        player_label.setWordWrap(True)
        parent_layout.addWidget(player_label, 2 + row_id, _PRESET_COLUMNS + 1 + team_id * 3)

        connection_state = None
        if not is_observer:
            connection_state = QtWidgets.QLabel(self.players_box)
            connection_state.setText("")
            connection_state.setWordWrap(True)
            parent_layout.addWidget(connection_state, 2 + row_id, _PRESET_COLUMNS + 2 + team_id * 3)

        tool_button = QtWidgets.QToolButton(self.players_box)
        tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        tool_button.setText("...")
        parent_layout.addWidget(tool_button, 2 + row_id, _PRESET_COLUMNS + 3 + team_id * 3)

        tool_button_menu = QtWidgets.QMenu(tool_button)
        tool_button.setMenu(tool_button_menu)

        widget = PlayerWidget(
            name=player_label,
            connection_state=connection_state,
            tool=tool_button,
            kick=QtWidgets.QAction(tool_button_menu),
            promote=QtWidgets.QAction(tool_button_menu),
            open_tracker=QtWidgets.QAction(tool_button_menu),
            abandon=QtWidgets.QAction(tool_button_menu),
            move_up=QtWidgets.QAction(tool_button_menu),
            move_down=QtWidgets.QAction(tool_button_menu),
            switch_observer_action=QtWidgets.QAction(tool_button_menu),
        )

        def player_action(action: SessionAdminUserAction, arg):
            return functools.partial(self._admin_player_action_slot, widget, action, arg)

        widget.kick.setText("Kick Player")
        widget.kick.triggered.connect(player_action(SessionAdminUserAction.KICK, None))
        tool_button_menu.addAction(widget.kick)

        widget.promote.setText("Promote to Admin")
        widget.promote.triggered.connect(player_action(SessionAdminUserAction.SWITCH_ADMIN, None))
        tool_button_menu.addAction(widget.promote)

        widget.open_tracker.setText("Open Tracker (NYI)")
        tool_button_menu.addAction(widget.open_tracker)

        widget.abandon.setText("Abandon (NYI)")
        widget.abandon.triggered.connect(player_action(SessionAdminUserAction.ABANDON, None))
        tool_button_menu.addAction(widget.abandon)

        widget.move_up.setText("Move up one slot")
        widget.move_up.triggered.connect(player_action(SessionAdminUserAction.MOVE, -1))
        tool_button_menu.addAction(widget.move_up)

        widget.move_down.setText("Move down one slot")
        widget.move_down.triggered.connect(player_action(SessionAdminUserAction.MOVE, 1))
        tool_button_menu.addAction(widget.move_down)

        widget.switch_observer_action.setText("Move to Observer")
        widget.switch_observer_action.triggered.connect(functools.partial(self._switch_observer_action, widget))
        tool_button_menu.addAction(widget.switch_observer_action)

        if is_observer:
            self.observers.append(widget)
        else:
            self.team_players.append(widget)

    @asyncSlot(GameSessionEntry)
    async def on_game_session_updated(self, game_session: GameSessionEntry):
        self._game_session = game_session

        if self.network_client.current_user.id not in game_session.players:
            return await self._on_kicked()

        self.session_name_label.setText(game_session.name)
        self.advanced_options_tool.setEnabled(game_session.players[self.network_client.current_user.id].admin)

        self.sync_rows_to_game_session()
        self.sync_player_widgets_to_game_session()
        self.sync_background_process_to_game_session()
        self.update_game_tab()
        self.update_session_actions()
        await self.update_multiworld_client_status()
        await self.update_logic_settings_window()

    async def _on_kicked(self):
        if self._already_kicked:
            return
        self._already_kicked = True
        leave_session = self.network_client.leave_game_session(False)
        if not self._expecting_kick:
            if self._game_session.players:
                message = "Kicked", "You have been kicked out of the session."
            else:
                message = "Session deleted", "The session has been deleted."
            await asyncio.gather(async_dialog.warning(self, *message), leave_session)
        else:
            await leave_session
        return QTimer.singleShot(0, self.close)

    def sync_rows_to_game_session(self):
        game_session = self._game_session
        self_player = game_session.players[self.network_client.current_user.id]

        while len(self.rows) > len(game_session.presets):
            self.pop_row()

        while len(self.rows) < len(game_session.presets):
            self.add_row()

        can_customize = game_session.generation_in_progress is None and game_session.seed_hash is None
        for i, (row, preset) in enumerate(zip(self.rows, game_session.presets)):
            row.set_preset(preset)
            row.set_is_admin(self_player.admin, is_your_row=self_player.row == i, can_customize=can_customize)

    def sync_player_widgets_to_game_session(self):
        game_session = self._game_session
        self_player = game_session.players[self.network_client.current_user.id]

        session_team = {}
        observers = []
        for player in game_session.players.values():
            if player.is_observer:
                observers.append(player)
            else:
                session_team[player.row] = player

        for row_id, player_widget in enumerate(self.team_players):
            player_widget.set_player(session_team.get(row_id))
            player_widget.update_player_widget(game_session, self_player)

        while len(self.observers) > len(observers):
            self.observers.pop().delete_widgets()

        while len(self.observers) < len(observers):
            self.append_new_player_widget(True)

        for observer, observer_widget in zip(observers, self.observers):
            observer_widget.set_player(observer)
            observer_widget.update_player_widget(game_session, self_player)

    def sync_background_process_to_game_session(self):
        game_session = self._game_session
        if game_session.generation_in_progress is not None:
            if not self._generating_game:
                other_player = game_session.players[game_session.generation_in_progress]
                self.progress_label.setText(f"Game being generated by {other_player.name}")

        elif self.has_background_process:
            if self._generating_game or game_session.seed_hash is None:
                self.stop_background_process()
        else:
            self.progress_label.setText("")

    def update_game_tab(self):
        game_session = self._game_session
        self_is_admin = game_session.players[self.network_client.current_user.id].admin

        self.session_status_label.setText(f"Session: {game_session.state.user_friendly_name}")
        self.update_background_process_button()
        self.generate_game_with_spoiler_action.setEnabled(self_is_admin)
        self.generate_game_without_spoiler_action.setEnabled(self_is_admin)
        self.import_permalink_action.setEnabled(self_is_admin)
        self.session_status_tool.setEnabled(self_is_admin)
        _state_to_label = {
            GameSessionState.SETUP: "Start",
            GameSessionState.IN_PROGRESS: "Finish",
            GameSessionState.FINISHED: "Reset",
        }
        self.session_status_tool.setText(_state_to_label[game_session.state])

        self.save_iso_button.setEnabled(game_session.seed_hash is not None
                                        and not self.current_player_membership.is_observer)
        if game_session.seed_hash is None:
            self.generate_game_label.setText("<Game not generated>")
            self.view_game_details_button.setEnabled(False)
        else:
            self.generate_game_label.setText(f"Seed hash: {game_session.word_hash} ({game_session.seed_hash})")
            self.view_game_details_button.setEnabled(game_session.spoiler)

        self.start_session_action.setEnabled(self_is_admin and game_session.state == GameSessionState.SETUP)
        self.finish_session_action.setEnabled(self_is_admin and game_session.state == GameSessionState.IN_PROGRESS)
        self.reset_session_action.setEnabled(self_is_admin and game_session.state != GameSessionState.SETUP)

    def update_session_actions(self):
        scrollbar = self.history_table_widget.verticalScrollBar()
        autoscroll = scrollbar.value() == scrollbar.maximum()
        self.history_table_widget.horizontalHeader().setVisible(True)
        self.history_table_widget.setRowCount(len(self._game_session.actions))
        for i, action in enumerate(self._game_session.actions):
            self.history_table_widget.setItem(i, 0, QtWidgets.QTableWidgetItem(action.message))
            self.history_table_widget.setItem(i, 1, QtWidgets.QTableWidgetItem(action.time.astimezone().strftime("%c")))
        if autoscroll:
            self.history_table_widget.scrollToBottom()

    async def update_multiworld_client_status(self):
        game_session_in_progress = self._game_session.state == GameSessionState.IN_PROGRESS

        if game_session_in_progress:
            if not self.multiworld_client.is_active:
                you = self._game_session.players[self.network_client.current_user.id]
                session_id = f"{self._game_session.id}_{self._game_session.seed_hash}_{you.row}"
                persist_path = self.network_client.server_data_path.joinpath(f"game_session_{session_id}.json")
                await self.multiworld_client.start(persist_path)

        elif self.multiworld_client.is_active:
            await self.multiworld_client.stop()

    async def update_logic_settings_window(self):
        if self._logic_settings_window is not None:
            if self._game_session.seed_hash is not None:
                self._logic_settings_window.reject()
                await async_dialog.warning(self, "Game was generated",
                                           "A game was generated, so changing presets is no longer possible.")
            else:
                self._logic_settings_window.setEnabled(self._game_session.generation_in_progress is None)

    @property
    def current_player_membership(self) -> PlayerSessionEntry:
        user = self.network_client.current_user
        return self._game_session.players[user.id]

    @asyncSlot()
    @handle_network_errors
    async def _admin_global_action_slot(self, action: SessionAdminGlobalAction, arg):
        return await self._admin_global_action(action, arg)

    async def _admin_global_action(self, action: SessionAdminGlobalAction, arg):
        self.setEnabled(False)
        try:
            return await self.network_client.session_admin_global(action, arg)
        finally:
            self.setEnabled(True)

    @asyncSlot()
    @handle_network_errors
    async def _admin_player_action_slot(self, player: PlayerWidget, action: SessionAdminUserAction, arg):
        if player.player is None:
            raise RuntimeError("Admin action attempted on empty slot")
        return await self._admin_player_action(player.player, action, arg)

    async def _admin_player_action(self, player: PlayerSessionEntry, action: SessionAdminUserAction, arg):
        self.setEnabled(False)
        try:
            return await self.network_client.session_admin_player(player.id, action, arg)
        finally:
            self.setEnabled(True)

    @asyncSlot()
    @handle_network_errors
    async def _switch_observer_action(self, widget: PlayerWidget):
        if widget.player.is_observer:
            if any(row.player is None for row in self.team_players):
                return await self._admin_player_action(widget.player, SessionAdminUserAction.SWITCH_IS_OBSERVER, 0)
            else:
                return await async_dialog.warning(self, "No free slot",
                                                  "There are no free slots for players in this session.\n\n"
                                                  "Press 'New Row' to add more if needed.")
        else:
            return await self._admin_player_action(widget.player, SessionAdminUserAction.SWITCH_IS_OBSERVER, None)

    @asyncSlot()
    @handle_network_errors
    async def rename_session(self):
        await async_dialog.warning(self, "Not yet implemented",
                                   "Renaming session isn't implemented yet.")

    @asyncSlot()
    @handle_network_errors
    async def change_password(self):
        dialog = QtWidgets.QInputDialog(self)
        dialog.setModal(True)
        dialog.setWindowTitle("Enter password")
        dialog.setLabelText("Enter the new password for the session:")
        dialog.setTextEchoMode(QtWidgets.QLineEdit.Password)
        if await async_dialog.execute_dialog(dialog) == QtWidgets.QDialog.Accepted:
            await self._admin_global_action(SessionAdminGlobalAction.CHANGE_PASSWORD, dialog.textValue())

    @asyncSlot()
    @handle_network_errors
    async def delete_session(self):
        result = await async_dialog.warning(self, "Delete session",
                                            "Are you sure you want to delete this session?",
                                            QMessageBox.Yes | QMessageBox.No,
                                            QMessageBox.No)
        if result == QMessageBox.Yes:
            self._expecting_kick = True
            try:
                await self._admin_global_action(SessionAdminGlobalAction.DELETE_SESSION, None)
            except Exception:
                self._expecting_kick = False
                raise

    async def _check_dangerous_presets(self, permalink: Permalink) -> bool:
        all_dangerous_settings = {}
        for i, preset in permalink.presets.items():
            dangerous = preset.dangerous_settings()
            if dangerous:
                all_dangerous_settings[i] = dangerous

        if all_dangerous_settings:
            player_names = {
                i: widget.player.name
                for i, widget in enumerate(self.team_players)
                if widget.player is not None
            }

            warnings = "\n".join(
                f"{player_names.get(i, f'Player {i + 1}')} - {self._game_session.presets[i].name}: {', '.join(dangerous)}"
                for i, dangerous in all_dangerous_settings.items()
            )
            message = ("The following presets have settings that can cause an impossible game:\n"
                       f"\n{warnings}\n"
                       "\nDo you want to continue?")
            result = await async_dialog.warning(self, "Dangerous preset", message, QMessageBox.Yes | QMessageBox.No)
            if result == QMessageBox.No:
                return False

        return True

    async def generate_game(self, spoiler: bool):
        if not self._options.is_alert_displayed(InfoAlert.MULTI_ENERGY_ALERT):
            await async_dialog.warning(self, "Multiworld Limitation",
                                       "Warning: Multiworld games doesn't have proper energy damage logic. "
                                       "You might be required to do Dark Aether checks with very low energy.")
            self._options.mark_alert_as_displayed(InfoAlert.MULTI_ENERGY_ALERT)

        permalink = Permalink(
            seed_number=random.randint(0, 2 ** 31),
            spoiler=spoiler,
            presets={
                i: preset.get_preset()
                for i, preset in enumerate(self._game_session.presets)
            },
        )
        return await self.generate_game_with_permalink(permalink)

    async def generate_game_with_permalink(self, permalink: Permalink):
        if not await self._check_dangerous_presets(permalink):
            return

        def generate_layout(progress_update: ProgressUpdateCallable):
            return simplified_patcher.generate_layout(progress_update=progress_update,
                                                      permalink=permalink,
                                                      options=self._options)

        await self._admin_global_action(SessionAdminGlobalAction.UPDATE_LAYOUT_GENERATION, True)
        self._generating_game = True
        try:
            layout = await self.run_in_background_async(generate_layout, "Creating a seed...")
            self.update_progress(f"Finished generating, uploading...", 100)
            await self._upload_layout_description(layout)
            self.update_progress("Uploaded!", 100)

        except Exception as e:
            await self._admin_global_action(SessionAdminGlobalAction.UPDATE_LAYOUT_GENERATION, False)

            message = "Error"
            if isinstance(e, GenerationFailure):
                message = "Generation Failure"
                self.failure_handler.handle_failure(e)
            else:
                logger.exception("Unable to generate")

            self.update_progress(f"{message}: {e}", -1)

        finally:
            self._generating_game = False

    async def clear_generated_game(self):
        result = await async_dialog.warning(
            self, "Clear generated game?",
            "Clearing the generated game will allow presets to be customized again, but all "
            "players must export the ISOs again.",
            QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
        if result == QMessageBox.Yes:
            await self._admin_global_action(SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION, None)

    @asyncSlot()
    @handle_network_errors
    async def generate_game_with_spoiler(self):
        await self.generate_game(True)

    @asyncSlot()
    @handle_network_errors
    async def generate_game_without_spoiler(self):
        await async_dialog.warning(self, "Not yet implemented",
                                   "Online game sessions without spoilers aren't available right now.")

    @asyncSlot()
    @handle_network_errors
    async def import_permalink(self):
        dialog = PermalinkDialog()
        result = await async_dialog.execute_dialog(dialog)
        if result != QtWidgets.QDialog.Accepted:
            return

        permalink = dialog.get_permalink_from_field()
        if permalink.player_count != self._game_session.num_rows:
            return await async_dialog.warning(
                self, "Incompatible permalink",
                f"Given permalink is for {permalink.player_count} players, but "
                f"this session only have {self._game_session.num_rows} rows.")

        if any(not preset_p.is_same_configuration(preset_s.get_preset())
               for preset_p, preset_s in zip(permalink.presets.values(), self._game_session.presets)):
            response = await async_dialog.warning(
                self, "Different presets",
                f"Given permalink has different presets compared to the session.\n"
                f"Do you want to overwrite the session's presets?",
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No
            )
            if response != QMessageBox.Yes:
                return

        await self.generate_game_with_permalink(permalink)

    async def _upload_layout_description(self, layout: LayoutDescription):
        await self._admin_global_action(SessionAdminGlobalAction.CHANGE_LAYOUT_DESCRIPTION,
                                        layout.as_json)

    @asyncSlot()
    @handle_network_errors
    async def start_session(self):
        num_players = sum(1 for player in self._game_session.players.values() if not player.is_observer)
        if num_players != self._game_session.num_rows:
            await async_dialog.message_box(self,
                                           QMessageBox.Critical,
                                           "Missing players",
                                           "Unable to start the session: there are missing players.\n"
                                           "All slots of all teams must be filled before start.",
                                           QMessageBox.Ok)
            return

        await self._admin_global_action(SessionAdminGlobalAction.START_SESSION, None)

    @asyncSlot()
    @handle_network_errors
    async def finish_session(self):
        result = await async_dialog.warning(
            self, "Finish session?",
            "It's no longer possible to collect items after the session is finished."
            "\nDo you want to continue?",
            QMessageBox.Yes | QMessageBox.No,
            QMessageBox.No,
        )
        if result == QMessageBox.Yes:
            await self._admin_global_action(SessionAdminGlobalAction.FINISH_SESSION, None)

    @asyncSlot()
    @handle_network_errors
    async def reset_session(self):
        await async_dialog.warning(self, "NYI", "Reset session is not implemented.")

    @asyncSlot()
    async def _session_status_button_clicked(self):
        state = self._game_session.state
        if state == GameSessionState.SETUP:
            await self.start_session()
        elif state == GameSessionState.IN_PROGRESS:
            await self.finish_session()
        elif state == GameSessionState.FINISHED:
            await self.reset_session()
        else:
            raise RuntimeError(f"Unknown session state: {state}")

    @asyncSlot()
    @handle_network_errors
    async def save_iso(self):
        membership = self.current_player_membership
        if membership.is_observer:
            return await async_dialog.message_box(self, QtWidgets.QMessageBox.Critical,
                                                  "Invalid action", "Observers can't generate an ISO.", QMessageBox.Ok)

        if any(player.player is None for player in self.team_players):
            user_response = await async_dialog.warning(
                self,
                "Incomplete Session",
                ("This session is currently missing a player.\n"
                 "If you create an ISO right now, all references to that player will use a generic name instead.\n\n"
                 "Do you want to proceed?"),
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No
            )
            if user_response == QMessageBox.No:
                return

        options = self._options

        if not options.is_alert_displayed(InfoAlert.MULTIWORLD_FAQ):
            await async_dialog.message_box(self, QMessageBox.Icon.Information, "Multiworld FAQ",
                                           "Have you read the Multiworld FAQ?\n"
                                           "It can be found in the main Randovania window → Help → Multiworld")
            options.mark_alert_as_displayed(InfoAlert.MULTIWORLD_FAQ)

        dialog = GameInputDialog(options, "Echoes Randomizer - {}.iso".format(self._game_session.word_hash), False)
        result = await async_dialog.execute_dialog(dialog)

        if result != QtWidgets.QDialog.Accepted:
            return

        patcher_data = await self._admin_player_action(membership, SessionAdminUserAction.CREATE_PATCHER_FILE,
                                                       options.cosmetic_patches.as_json)
        shareable_hash = self._game_session.seed_hash

        configuration = self._game_session.presets[membership.row].get_preset().configuration
        game = data_reader.decode_data(configuration.game_data)
        game_specific = base_patches_factory.create_game_specific(configuration, game)

        input_file = dialog.input_file
        output_file = dialog.output_file

        with options:
            options.output_directory = output_file.parent

        def work(progress_update: ProgressUpdateCallable):
            num_updaters = 2
            if input_file is not None:
                num_updaters += 1

            updaters = status_update_lib.split_progress_update(progress_update, num_updaters)
            if input_file is not None:
                simplified_patcher.unpack_iso(input_iso=input_file,
                                              options=options,
                                              progress_update=updaters[0])

            # Apply Layout
            simplified_patcher.apply_patcher_file(patcher_file=patcher_data,
                                                  game_specific=game_specific,
                                                  shareable_hash=shareable_hash,
                                                  options=options,
                                                  progress_update=updaters[-2])

            # Pack ISO
            simplified_patcher.pack_iso(output_iso=output_file,
                                        options=options,
                                        progress_update=updaters[-1])
            progress_update(f"Finished!", 1)

        self.run_in_background_thread(work, "Exporting...")

    @asyncSlot()
    @handle_network_errors
    async def view_game_details(self):
        if self._game_session.seed_hash is None:
            return await async_dialog.warning(self, "No Spoiler Available",
                                              "Unable to view game spoilers, no game available.")

        if not self._game_session.spoiler:
            return await async_dialog.warning(self, "No Spoiler Available",
                                              "Unable to view game spoilers, game was generated without spoiler.")

        description_json = await self._admin_global_action(SessionAdminGlobalAction.DOWNLOAD_LAYOUT_DESCRIPTION, None)
        description = LayoutDescription.from_json_dict(json.loads(description_json))
        self._window_manager.open_game_details(description)

    @asyncSlot()
    async def copy_permalink(self):
        permalink_str = self._game_session.permalink
        dialog = QtWidgets.QInputDialog(self)
        dialog.setModal(True)
        dialog.setWindowTitle("Session permalink")
        dialog.setLabelText("Permalink:")
        dialog.setTextValue(permalink_str)
        QtWidgets.QApplication.clipboard().setText(permalink_str)
        await async_dialog.execute_dialog(dialog)

    @asyncSlot()
    @handle_network_errors
    async def background_process_button_clicked(self):
        if self.has_background_process:
            self.stop_background_process()
        elif self._game_session.generation_in_progress is not None:
            await self._admin_global_action(SessionAdminGlobalAction.UPDATE_LAYOUT_GENERATION, False)
        elif self._game_session.seed_hash is not None:
            await self.clear_generated_game()
        else:
            await self.generate_game(True)

    def update_background_process_button(self):
        is_admin = self.current_player_membership.admin
        if self.has_background_process:
            self.background_process_button.setEnabled(self.has_background_process or is_admin)
            self.background_process_button.setText("Stop")

        elif self._game_session.generation_in_progress is not None:
            self.background_process_button.setEnabled(is_admin)
            self.background_process_button.setText("Abort generation")
        else:
            self.background_process_button.setEnabled(self._game_session.state == GameSessionState.SETUP and is_admin)
            if self._game_session.seed_hash is not None:
                self.background_process_button.setText("Clear generated game")
            else:
                self.background_process_button.setText("Generate game")

    def enable_buttons_with_background_tasks(self, value: bool):
        self.update_background_process_button()

    def update_progress(self, message: str, percentage: int):
        self.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0
        if percentage >= 0:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(percentage)
        else:
            self.progress_bar.setRange(0, 0)

    def _open_user_preferences_dialog(self):
        dialog = EchoesUserPreferencesDialog(self, self._options.cosmetic_patches)
        result = dialog.exec_()
        if result == QtWidgets.QDialog.Accepted:
            with self._options as options:
                options.cosmetic_patches = dialog.cosmetic_patches

    def on_server_connection_state_updated(self, state: ConnectionState):
        self.server_connection_button.setEnabled(state == ConnectionState.Disconnected)
        self.server_connection_label.setText(f"Server: {state.value}")

    @asyncSlot()
    @handle_network_errors
    async def _connect_to_server(self):
        await self.network_client.connect_to_server()

    @asyncSlot()
    async def on_game_connection_updated(self):
        async with self._update_status_lock:
            await self.network_client.session_self_update(self.game_connection.get_current_inventory(),
                                                          self.game_connection.current_status,
                                                          self.game_connection.backend.backend_choice)



import json
import sys
from pathlib import Path
from typing import Optional

from randovania.version import version

CONFIGURATION_FILE_PATH: Optional[Path] = None


def is_frozen() -> bool:
    return getattr(sys, "frozen", False)


def get_data_path() -> Path:
    if is_frozen():
        file_dir = Path(getattr(sys, "_MEIPASS"))
    else:
        file_dir = Path(__file__).parent
    return file_dir.joinpath("data")


def _get_default_configuration_path() -> Path:
    return get_data_path().joinpath("configuration.json")


def get_configuration() -> dict:
    file_path = CONFIGURATION_FILE_PATH
    if file_path is None:
        file_path = _get_default_configuration_path()

    try:
        with file_path.open() as file:
            return json.load(file)
    except FileNotFoundError:
        if CONFIGURATION_FILE_PATH is None:
            return {
                "server_address": "http://127.0.0.1:5000",
                "socketio_path": "/socket.io",
            }
        else:
            raise


__version__ = version
VERSION = version



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_translators.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetEchoesTranslators(object):
    def setupUi(self, PresetEchoesTranslators):
        if not PresetEchoesTranslators.objectName():
            PresetEchoesTranslators.setObjectName(u"PresetEchoesTranslators")
        PresetEchoesTranslators.resize(466, 454)
        self.centralWidget = QWidget(PresetEchoesTranslators)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.translators_top_layout = QGridLayout(self.centralWidget)
        self.translators_top_layout.setSpacing(6)
        self.translators_top_layout.setContentsMargins(11, 11, 11, 11)
        self.translators_top_layout.setObjectName(u"translators_top_layout")
        self.translators_top_layout.setContentsMargins(0, 4, 0, 0)
        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setFrameShadow(QFrame.Plain)
        self.scroll_area.setLineWidth(0)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.translators_scroll_contents = QWidget()
        self.translators_scroll_contents.setObjectName(u"translators_scroll_contents")
        self.translators_scroll_contents.setGeometry(QRect(0, 0, 466, 450))
        self.translators_layout = QGridLayout(self.translators_scroll_contents)
        self.translators_layout.setSpacing(6)
        self.translators_layout.setContentsMargins(11, 11, 11, 11)
        self.translators_layout.setObjectName(u"translators_layout")
        self.translators_layout.setContentsMargins(0, 0, 0, 0)
        self.translator_randomize_all_button = QPushButton(self.translators_scroll_contents)
        self.translator_randomize_all_button.setObjectName(u"translator_randomize_all_button")

        self.translators_layout.addWidget(self.translator_randomize_all_button, 1, 0, 1, 1)

        self.translator_vanilla_colors_button = QPushButton(self.translators_scroll_contents)
        self.translator_vanilla_colors_button.setObjectName(u"translator_vanilla_colors_button")

        self.translators_layout.addWidget(self.translator_vanilla_colors_button, 1, 2, 1, 1)

        self.translator_vanilla_actual_button = QPushButton(self.translators_scroll_contents)
        self.translator_vanilla_actual_button.setObjectName(u"translator_vanilla_actual_button")

        self.translators_layout.addWidget(self.translator_vanilla_actual_button, 1, 1, 1, 1)

        self.translators_description = QLabel(self.translators_scroll_contents)
        self.translators_description.setObjectName(u"translators_description")
        self.translators_description.setWordWrap(True)

        self.translators_layout.addWidget(self.translators_description, 0, 0, 1, 3)

        self.scroll_area.setWidget(self.translators_scroll_contents)

        self.translators_top_layout.addWidget(self.scroll_area, 0, 0, 1, 1)

        PresetEchoesTranslators.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesTranslators)

        QMetaObject.connectSlotsByName(PresetEchoesTranslators)
    # setupUi

    def retranslateUi(self, PresetEchoesTranslators):
        PresetEchoesTranslators.setWindowTitle(QCoreApplication.translate("PresetEchoesTranslators", u"Translators Gate", None))
        self.translator_randomize_all_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Randomize All", None))
        self.translator_vanilla_colors_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Vanilla (Colors)", None))
        self.translator_vanilla_actual_button.setText(QCoreApplication.translate("PresetEchoesTranslators", u"Vanilla (Actual)", None))
        self.translators_description.setText(QCoreApplication.translate("PresetEchoesTranslators", u"<html><head/><body><p>Change which translator is required for all the gates in the game. Their colors are changed to match the necessary translator.</p><p>There are two vanilla options: using the actual translator requirements in the game, and using the vanilla gate colors.<br/>This is because in the original game, some translator gates are colored one way, but the translator requirement is something else.</p><p>The Emerald gate in Great Temple - Temple Sanctuary has a special case, when elevators aren't vanilla: if you enter Great Temple via Transport A, it will be permanently down.</p></body></html>", None))
    # retranslateUi




import dataclasses
from functools import lru_cache
from math import ceil
from typing import NamedTuple, Optional, Iterable, FrozenSet, Iterator, Tuple, List, Type, Union

from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo, CurrentResources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo

MAX_DAMAGE = 9999999


class Requirement:
    def damage(self, current_resources: CurrentResources) -> int:
        raise NotImplementedError()

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        raise NotImplementedError()

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float,
                           ) -> "Requirement":
        """
        Creates a new Requirement that does not contain reference to resources in static_resources.
        For those that contains a reference, they're replaced with Trivial when satisfied and Impossible otherwise.
        :param static_resources:
        :param damage_multiplier: All damage requirements have their value multiplied by this.
        """
        raise NotImplementedError()

    def simplify(self) -> "Requirement":
        """
        Creates a new Requirement without some redundant complexities, like:
        - RequirementAnd/RequirementOr of exactly one item
        - RequirementAnd/RequirementOr of others of the same type.
        - RequirementAnd with impossible among the items
        - RequirementOr with trivial among the items
        :return:
        """
        raise NotImplementedError()

    @property
    def as_set(self) -> "RequirementSet":
        raise NotImplementedError()

    @classmethod
    @lru_cache()
    def trivial(cls) -> "Requirement":
        # empty RequirementAnd.satisfied is True
        return RequirementAnd([])

    @classmethod
    @lru_cache()
    def impossible(cls) -> "Requirement":
        # empty RequirementOr.satisfied is False
        return RequirementOr([])

    def __lt__(self, other: "Requirement"):
        return str(self) < str(other)

    def iterate_resource_requirements(self):
        raise NotImplementedError()


class RequirementAnd(Requirement):
    items: Tuple[Requirement, ...]
    _cached_hash = None

    def __init__(self, items: Iterable[Requirement]):
        self.items = tuple(items)

    def damage(self, current_resources: CurrentResources) -> int:
        result = 0
        for item in self.items:
            if item.satisfied(current_resources, MAX_DAMAGE):
                result += item.damage(current_resources)
            else:
                return MAX_DAMAGE
        return result

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        return all(
            item.satisfied(current_resources, current_energy)
            for item in self.items
        )

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float,
                           ) -> Requirement:
        return RequirementAnd(
            item.patch_requirements(static_resources, damage_multiplier) for item in self.items
        )

    def simplify(self) -> Requirement:
        new_items = _expand_items(self.items, RequirementAnd, Requirement.trivial())
        if Requirement.impossible() in new_items:
            return Requirement.impossible()

        if len(new_items) == 1:
            return new_items[0]

        return RequirementAnd(new_items)

    @property
    def as_set(self) -> "RequirementSet":
        result = RequirementSet.trivial()
        for item in self.items:
            result = result.union(item.as_set)
        return result

    @property
    def sorted(self) -> Tuple[Requirement]:
        return tuple(sorted(self.items))

    def __eq__(self, other):
        return isinstance(other, RequirementAnd) and self.items == other.items

    def __hash__(self) -> int:
        if self._cached_hash is None:
            self._cached_hash = hash(self.items)
        return self._cached_hash

    def __repr__(self):
        return repr(self.items)

    def __str__(self) -> str:
        if self.items:
            visual_items = [str(item) for item in self.items]
            return "({})".format(" and ".join(sorted(visual_items)))
        else:
            return "Trivial"

    def iterate_resource_requirements(self):
        for item in self.items:
            yield from item.iterate_resource_requirements()


class RequirementOr(Requirement):
    items: Tuple[Requirement, ...]
    _cached_hash = None

    def __init__(self, items: Iterable[Requirement]):
        self.items = tuple(items)

    def damage(self, current_resources: CurrentResources) -> int:
        try:
            return min(
                item.damage(current_resources)
                for item in self.items
                if item.satisfied(current_resources, MAX_DAMAGE)
            )
        except ValueError:
            return MAX_DAMAGE

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        return any(
            item.satisfied(current_resources, current_energy)
            for item in self.items
        )

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float,
                           ) -> Requirement:
        return RequirementOr(
            item.patch_requirements(static_resources, damage_multiplier) for item in self.items
        )

    def simplify(self) -> Requirement:
        new_items = _expand_items(self.items, RequirementOr, Requirement.impossible())
        if Requirement.trivial() in new_items:
            return Requirement.trivial()

        num_and_requirements = 0
        common_requirements = None
        for item in new_items:
            if isinstance(item, RequirementAnd):
                num_and_requirements += 1
                if common_requirements is None:
                    common_requirements = item.items
                else:
                    common_requirements = [
                        common
                        for common in common_requirements
                        if common in item.items
                    ]

        # Only extract the common requirements if there's more than 1 requirement
        if num_and_requirements >= 2 and common_requirements:
            simplified_items = []
            common_new_or = []

            for item in new_items:
                if isinstance(item, RequirementAnd):
                    assert set(common_requirements) <= set(item.items)
                    simplified_condition = [it for it in item.items if it not in common_requirements]
                    if simplified_condition:
                        common_new_or.append(RequirementAnd(simplified_condition) if len(simplified_condition) > 1
                                             else simplified_condition[0])
                else:
                    simplified_items.append(item)

            common_requirements.append(RequirementOr(common_new_or))
            simplified_items.append(RequirementAnd(common_requirements))
            final_items = simplified_items

        else:
            final_items = new_items

        if len(final_items) == 1:
            return final_items[0]

        return RequirementOr(final_items)

    @property
    def as_set(self) -> "RequirementSet":
        alternatives = set()
        for item in self.items:
            alternatives |= item.as_set.alternatives
        return RequirementSet(alternatives)

    @property
    def sorted(self) -> Tuple[Requirement]:
        return tuple(sorted(self.items))

    def __eq__(self, other):
        return isinstance(other, RequirementOr) and self.items == other.items

    def __hash__(self) -> int:
        if self._cached_hash is None:
            self._cached_hash = hash(self.items)
        return self._cached_hash

    def __repr__(self):
        return repr(self.items)

    def __str__(self) -> str:
        if self.items:
            visual_items = [str(item) for item in self.items]
            return "({})".format(" or ".join(sorted(visual_items)))
        else:
            return "Impossible"

    def iterate_resource_requirements(self):
        for item in self.items:
            yield from item.iterate_resource_requirements()


def _expand_items(items: Tuple[Requirement, ...],
                  cls: Type[Union[RequirementAnd, RequirementOr]],
                  exclude: Requirement) -> List[Requirement]:
    expanded = []

    def _add(_item):
        if _item not in expanded and _item != exclude:
            expanded.append(_item)

    for item in items:
        simplified = item.simplify()
        if isinstance(simplified, cls):
            for new_item in simplified.items:
                _add(new_item)
        else:
            _add(simplified)
    return expanded


@dataclasses.dataclass(frozen=True)
class ResourceRequirement(Requirement):
    resource: ResourceInfo
    amount: int
    negate: bool

    @classmethod
    def with_data(cls,
                  database: ResourceDatabase,
                  resource_type: ResourceType,
                  requirement_index: int,
                  amount: int,
                  negate: bool) -> "ResourceRequirement":
        return cls(
            database.get_by_type_and_index(resource_type, requirement_index),
            amount,
            negate)

    @property
    def is_damage(self) -> bool:
        return self.resource.resource_type == ResourceType.DAMAGE

    def damage(self, current_resources: CurrentResources) -> int:
        if self.resource.resource_type == ResourceType.DAMAGE:
            return ceil(self.resource.damage_reduction(current_resources) * self.amount)
        else:
            return 0

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        """Checks if a given resources dict satisfies this requirement"""

        if self.is_damage:
            assert not self.negate, "Damage requirements shouldn't have the negate flag"

            return current_energy > self.damage(current_resources)

        has_amount = current_resources.get(self.resource, 0) >= self.amount
        if self.negate:
            return not has_amount
        else:
            return has_amount

    def simplify(self) -> Requirement:
        return self

    def __repr__(self):
        return "{} {} {}".format(
            self.resource,
            "<" if self.negate else "≥",
            self.amount)

    @property
    def pretty_text(self):
        if self.amount == 1:
            negated_prefix = self.resource.resource_type.negated_prefix
            non_negated_prefix = self.resource.resource_type.non_negated_prefix
            return "{}{}".format(negated_prefix if self.negate else non_negated_prefix, self.resource)
        else:
            return str(self)

    @property
    def _as_comparison_tuple(self):
        return self.resource.resource_type, self.resource.index, self.amount, self.negate

    def __lt__(self, other: "ResourceRequirement") -> bool:
        return self._as_comparison_tuple < other._as_comparison_tuple

    def multiply_amount(self, multiplier: float) -> "ResourceRequirement":
        return ResourceRequirement(
            self.resource,
            self.amount * multiplier,
            self.negate,
        )

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float,
                           ) -> Requirement:
        if static_resources.get(self.resource) is not None:
            if self.satisfied(static_resources, 0):
                return Requirement.trivial()
            else:
                return Requirement.impossible()
        else:
            if self.is_damage:
                return self.multiply_amount(damage_multiplier)
            else:
                return self

    @property
    def as_set(self) -> "RequirementSet":
        return RequirementSet([
            RequirementList([
                self
            ])
        ])

    def iterate_resource_requirements(self):
        yield self


class RequirementTemplate(Requirement):
    database: ResourceDatabase
    template_name: str

    def __init__(self, database: ResourceDatabase, template_name: str):
        self.database = database
        self.template_name = template_name

    @property
    def template_requirement(self) -> Requirement:
        return self.database.requirement_template[self.template_name]

    def damage(self, current_resources: CurrentResources) -> int:
        return self.template_requirement.damage(current_resources)

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        return self.template_requirement.satisfied(current_resources, current_energy)

    def patch_requirements(self, static_resources: CurrentResources, damage_multiplier: float,
                           ) -> Requirement:
        return self.template_requirement.patch_requirements(static_resources, damage_multiplier)

    def simplify(self) -> Requirement:
        return self

    @property
    def as_set(self) -> "RequirementSet":
        return self.template_requirement.as_set

    def __eq__(self, other):
        return isinstance(other, RequirementTemplate) and self.template_name == other.template_name

    def __hash__(self) -> int:
        return hash(self.template_name)

    def __str__(self) -> str:
        return self.template_name

    def iterate_resource_requirements(self):
        yield from self.template_requirement.iterate_resource_requirements()


class RequirementList:
    items: FrozenSet[ResourceRequirement]
    _cached_hash: Optional[int] = None

    def __deepcopy__(self, memodict):
        return self

    def __init__(self, items: Iterable[ResourceRequirement]):
        self.items = frozenset(items)

    def __eq__(self, other):
        return isinstance(
            other, RequirementList) and self.items == other.items

    def __lt__(self, other: "RequirementList"):
        return self.items < other.items

    def __hash__(self) -> int:
        if self._cached_hash is None:
            self._cached_hash = hash(self.items)
        return self._cached_hash

    def __repr__(self):
        return repr(self.items)

    def __str__(self) -> str:
        if self.items:
            return ", ".join(map(str, sorted(self.items)))
        else:
            return "Trivial"

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        """
        A list is considered satisfied if each IndividualRequirement that belongs to it is satisfied.
        In particular, an empty RequirementList is considered satisfied.
        :param current_resources:
        :param current_energy:
        :return:
        """

        energy = current_energy
        for requirement in self.values():
            if requirement.satisfied(current_resources, current_energy):
                if requirement.resource.resource_type == ResourceType.DAMAGE:
                    energy -= requirement.damage(current_resources)
            else:
                return False
        return True

    def get(self, resource: ResourceInfo) -> Optional[ResourceRequirement]:
        """
        Gets an IndividualRequirement that uses the given resource
        :param resource:
        :return:
        """
        for item in self.values():
            if item.resource == resource:
                return item
        return None

    @property
    def dangerous_resources(self) -> Iterator[ResourceInfo]:
        """
        Return an iterator of all SimpleResourceInfo in this list that have the negate flag
        :return:
        """
        for individual in self.values():
            if individual.negate:
                yield individual.resource

    def values(self) -> FrozenSet[ResourceRequirement]:
        return self.items

    def union(self, other: "RequirementList") -> "RequirementList":
        return RequirementList(self.items | other.items)


class RequirementSet:
    """
    Represents multiple alternatives of satisfying a requirement.
    For example, going from A to B may be possible by having Grapple+Space Jump or Screw Attack.
    """
    alternatives: FrozenSet[RequirementList]
    _cached_hash: Optional[int] = None

    def __init__(self, alternatives: Iterable[RequirementList]):
        """
        Constructs a RequirementSet from given iterator of RequirementList.
        Redundant alternatives (Bombs or Bombs + Space Jump) are automatically removed.
        :param alternatives:
        """
        input_set = frozenset(alternatives)
        self.alternatives = frozenset(
            requirement
            for requirement in input_set
            if not any(other < requirement for other in input_set)
        )

    def __deepcopy__(self, memodict):
        return self

    def __eq__(self, other):
        return isinstance(
            other, RequirementSet) and self.alternatives == other.alternatives

    def __hash__(self) -> int:
        if self._cached_hash is None:
            self._cached_hash = hash(self.alternatives)
        return self._cached_hash

    def __repr__(self):
        return repr(self.alternatives)

    def pretty_print(self, indent="", print_function=print):
        to_print = []
        if self == RequirementSet.impossible():
            to_print.append("Impossible")
        elif self == RequirementSet.trivial():
            to_print.append("Trivial")
        else:
            to_print.extend(
                str(alternative)
                for alternative in self.alternatives
            )
        for line in sorted(to_print):
            print_function(indent + line)

    @classmethod
    @lru_cache()
    def trivial(cls) -> "RequirementSet":
        # empty RequirementList.satisfied is True
        return cls([RequirementList([])])

    @classmethod
    @lru_cache()
    def impossible(cls) -> "RequirementSet":
        # No alternatives makes satisfied always return False
        return cls([])

    def satisfied(self, current_resources: CurrentResources, current_energy: int) -> bool:
        """
        Checks if at least one alternative is satisfied with the given resources.
        In particular, an empty RequirementSet is *never* considered satisfied.
        :param current_resources:
        :param current_energy:
        :return:
        """
        return any(
            requirement_list.satisfied(current_resources, current_energy)
            for requirement_list in self.alternatives)

    def union(self, other: "RequirementSet") -> "RequirementSet":
        """Create a new RequirementSet that is only satisfied when both are satisfied"""
        return RequirementSet(
            a.union(b)
            for a in self.alternatives
            for b in other.alternatives)

    def expand_alternatives(self, other: "RequirementSet") -> "RequirementSet":
        """Create a new RequirementSet that is satisfied when either are satisfied."""
        return RequirementSet(self.alternatives | other.alternatives)

    @property
    def dangerous_resources(self) -> Iterator[ResourceInfo]:
        """
        Return an iterator of all SimpleResourceInfo in all alternatives that have the negate flag
        :return:
        """
        for alternative in self.alternatives:
            yield from alternative.dangerous_resources

    @property
    def all_individual(self) -> Iterator[ResourceRequirement]:
        """
        Iterates over all individual requirements involved in this set
        :return:
        """
        for alternative in self.alternatives:
            for individual in alternative.values():
                yield individual


SatisfiableRequirements = FrozenSet[RequirementList]



from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class RandovaniaGame(BitPackEnum, Enum):
    PRIME1 = "prime1"
    PRIME2 = "prime2"
    PRIME3 = "prime3"

    @property
    def short_name(self) -> str:
        return SHORT_NAME[self]

    @property
    def long_name(self) -> str:
        return LONG_NAME[self]


SHORT_NAME = {
    RandovaniaGame.PRIME1: "Prime",
    RandovaniaGame.PRIME2: "Echoes",
    RandovaniaGame.PRIME3: "Corruption"
}

LONG_NAME = {
    RandovaniaGame.PRIME1: "Metroid Prime",
    RandovaniaGame.PRIME2: "Metroid Prime 2: Echoes",
    RandovaniaGame.PRIME3: "Metroid Prime 3: Corruption"
}



import dataclasses
import json
from distutils.version import StrictVersion
from enum import Enum
from pathlib import Path
from typing import Optional, TypeVar, Callable, Any, Set, List

from randovania.game_connection.backend_choice import GameBackendChoice
from randovania.interface_common import persistence, update_checker
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.persisted_options import get_persisted_options_from_data, serialized_data_for_options

T = TypeVar("T")


def identity(v: T) -> T:
    return v


class InfoAlert(Enum):
    FAQ = "faq"
    MULTI_ENERGY_ALERT = "multi-energy-alert"
    MULTIWORLD_FAQ = "multi-faq"
    NINTENDONT_UNSTABLE = "nintendont-unstable"


@dataclasses.dataclass(frozen=True)
class Serializer:
    encode: Callable[[Any], Any]
    decode: Callable[[Any], Any]


def serialize_alerts(alerts: Set[InfoAlert]) -> List[str]:
    return sorted(a.value for a in alerts)


def decode_alerts(data: List[str]):
    result = set()
    for item in data:
        try:
            result.add(InfoAlert(item))
        except ValueError:
            continue

    return result


_SERIALIZER_FOR_FIELD = {
    "last_changelog_displayed": Serializer(identity, str),
    "advanced_validate_seed_after": Serializer(identity, bool),
    "advanced_timeout_during_generation": Serializer(identity, bool),
    "auto_save_spoiler": Serializer(identity, bool),
    "dark_mode": Serializer(identity, bool),
    "output_directory": Serializer(str, Path),
    "selected_preset_name": Serializer(identity, str),
    "cosmetic_patches": Serializer(lambda p: p.as_json, CosmeticPatches.from_json_dict),
    "displayed_alerts": Serializer(serialize_alerts, decode_alerts),
    "game_backend": Serializer(lambda it: it.value, GameBackendChoice),
    "nintendont_ip": Serializer(identity, str),
}


def _return_with_default(value: Optional[T], default_factory: Callable[[], T]) -> T:
    """
    Returns the given value is if it's not None, otherwise call default_factory
    :param value:
    :param default_factory:
    :return:
    """
    if value is None:
        return default_factory()
    else:
        return value


class DecodeFailedException(ValueError):
    pass


class Options:
    _data_dir: Path
    _on_options_changed: Optional[Callable[[], None]] = None
    _nested_autosave_level: int = 0
    _is_dirty: bool = False

    _last_changelog_displayed: str
    _advanced_validate_seed_after: Optional[bool] = None
    _advanced_timeout_during_generation: Optional[bool] = None
    _auto_save_spoiler: Optional[bool] = None
    _dark_mode: Optional[bool] = None
    _output_directory: Optional[Path] = None
    _selected_preset_name: Optional[str] = None
    _cosmetic_patches: Optional[CosmeticPatches] = None
    _displayed_alerts: Optional[Set[InfoAlert]] = None
    _game_backend: Optional[GameBackendChoice] = None
    _nintendont_ip: Optional[str] = None

    def __init__(self, data_dir: Path):
        self._data_dir = data_dir
        self._last_changelog_displayed = str(update_checker.strict_current_version())

    @classmethod
    def with_default_data_dir(cls) -> "Options":
        return cls(persistence.user_data_dir())

    def _read_persisted_options(self) -> Optional[dict]:
        try:
            with self._data_dir.joinpath("config.json").open() as options_file:
                return json.load(options_file)
        except FileNotFoundError:
            return None

    def _set_field(self, field_name: str, value):
        setattr(self, "_" + field_name, value)

    def load_from_disk(self, ignore_decode_errors: bool = False) -> bool:
        """
        Loads the file created with `_save_to_disk`.
        :param ignore_decode_errors: If True, errors in the config file are ignored.
        :return: True, if a valid file exists.
        """
        try:
            persisted_data = self._read_persisted_options()

        except json.decoder.JSONDecodeError as e:
            if ignore_decode_errors:
                persisted_data = None
            else:
                raise DecodeFailedException(f"Unable to decode JSON: {e}")

        if persisted_data is None:
            return False

        persisted_options = get_persisted_options_from_data(persisted_data)
        self.load_from_persisted(persisted_options, ignore_decode_errors)
        return True

    def load_from_persisted(self,
                            persisted: dict,
                            ignore_decode_errors: bool,
                            ):
        """
        Loads fields from the given persisted options.
        :param persisted:
        :param ignore_decode_errors:
        :return:
        """
        for field_name, serializer in _SERIALIZER_FOR_FIELD.items():
            value = persisted.get(field_name, None)
            if value is not None:
                try:
                    decoded = serializer.decode(value)
                except Exception as err:
                    if ignore_decode_errors:
                        decoded = None
                    else:
                        raise DecodeFailedException(
                            f"Unable to decode field {field_name}: {err}"
                        )

                if decoded is not None:
                    self._set_field(field_name, decoded)

    def _serialize_fields(self) -> dict:
        data_to_persist = {}
        for field_name, serializer in _SERIALIZER_FOR_FIELD.items():
            value = getattr(self, "_" + field_name, None)
            if value is not None:
                data_to_persist[field_name] = serializer.encode(value)

        return serialized_data_for_options(data_to_persist)

    def _save_to_disk(self):
        """Serializes the fields of this Option and writes then to a file."""
        self._is_dirty = False
        data_to_persist = self._serialize_fields()

        self._data_dir.mkdir(parents=True, exist_ok=True)

        # Write to a separate file, so we don't corrupt the existing one in case we unexpectedly
        # are unable to finish writing the file
        new_config_path = self._data_dir.joinpath("config_new.json")
        with new_config_path.open("w") as options_file:
            json.dump(data_to_persist, options_file,
                      indent=4, separators=(',', ': '))

        # Place the new, complete, config to the desired path
        config_path = self._data_dir.joinpath("config.json")
        new_config_path.replace(config_path)

    def __enter__(self):
        self._nested_autosave_level += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._nested_autosave_level == 1:
            if self._is_dirty:
                # TODO: maybe it should be an error to change options to different values in on_options_changed?
                # previous = self._serialize_fields()
                if self._on_options_changed is not None:
                    self._on_options_changed()
                # assert previous == self._serialize_fields()
                self._save_to_disk()
        self._nested_autosave_level -= 1

    # Events
    def _set_on_options_changed(self, value):
        self._on_options_changed = value

    on_options_changed = property(fset=_set_on_options_changed)

    # Reset

    def reset_to_defaults(self):
        self._check_editable_and_mark_dirty()
        self._advanced_validate_seed_after = None
        self._advanced_timeout_during_generation = None
        self._auto_save_spoiler = None
        self._cosmetic_patches = None
        self._displayed_alerts = None
        self._dark_mode = None
        self._game_backend = None
        self._nintendont_ip = None

    # Files paths
    @property
    def backup_files_path(self) -> Path:
        return self._data_dir.joinpath("backup")

    @property
    def game_files_path(self) -> Path:
        return self._data_dir.joinpath("extracted_game")

    @property
    def tracker_files_path(self) -> Path:
        return self._data_dir.joinpath("tracker")

    @property
    def data_dir(self) -> Path:
        return self._data_dir

    # Access to Direct fields
    @property
    def last_changelog_displayed(self) -> StrictVersion:
        return StrictVersion(self._last_changelog_displayed)

    @last_changelog_displayed.setter
    def last_changelog_displayed(self, value: StrictVersion):
        if value != self.last_changelog_displayed:
            self._check_editable_and_mark_dirty()
            self._last_changelog_displayed = str(value)

    @property
    def output_directory(self) -> Optional[Path]:
        return self._output_directory

    @output_directory.setter
    def output_directory(self, value: Optional[Path]):
        self._edit_field("output_directory", value)

    @property
    def auto_save_spoiler(self) -> bool:
        return _return_with_default(self._auto_save_spoiler, lambda: False)

    @auto_save_spoiler.setter
    def auto_save_spoiler(self, value: bool):
        self._edit_field("auto_save_spoiler", value)

    @property
    def dark_mode(self) -> bool:
        return _return_with_default(self._dark_mode, lambda: False)

    @dark_mode.setter
    def dark_mode(self, value: bool):
        self._edit_field("dark_mode", value)

    @property
    def selected_preset_name(self) -> Optional[str]:
        return self._selected_preset_name

    @selected_preset_name.setter
    def selected_preset_name(self, value: str):
        self._edit_field("selected_preset_name", value)

    @property
    def cosmetic_patches(self) -> CosmeticPatches:
        return _return_with_default(self._cosmetic_patches, CosmeticPatches.default)

    @cosmetic_patches.setter
    def cosmetic_patches(self, value: CosmeticPatches):
        self._edit_field("cosmetic_patches", value)

    @property
    def game_backend(self) -> GameBackendChoice:
        return _return_with_default(self._game_backend, lambda: GameBackendChoice.DOLPHIN)

    @game_backend.setter
    def game_backend(self, value: GameBackendChoice):
        self._edit_field("game_backend", value)

    @property
    def nintendont_ip(self) -> Optional[str]:
        return self._nintendont_ip

    @nintendont_ip.setter
    def nintendont_ip(self, value: Optional[str]):
        self._edit_field("nintendont_ip", value)

    @property
    def displayed_alerts(self):
        return _return_with_default(self._displayed_alerts, set)

    @displayed_alerts.setter
    def displayed_alerts(self, value):
        self._edit_field("displayed_alerts", value)

    def is_alert_displayed(self, value: InfoAlert):
        return value in self.displayed_alerts

    def mark_alert_as_displayed(self, value: InfoAlert):
        if value not in self.displayed_alerts:
            # Create a copy so we don't modify the existing field
            alerts = set(self.displayed_alerts)
            alerts.add(value)
            with self:
                self.displayed_alerts = alerts

    # Advanced

    @property
    def advanced_validate_seed_after(self) -> bool:
        return _return_with_default(self._advanced_validate_seed_after, lambda: True)

    @advanced_validate_seed_after.setter
    def advanced_validate_seed_after(self, value: bool):
        self._edit_field("advanced_validate_seed_after", value)

    @property
    def advanced_timeout_during_generation(self) -> bool:
        return _return_with_default(self._advanced_timeout_during_generation, lambda: True)

    @advanced_timeout_during_generation.setter
    def advanced_timeout_during_generation(self, value: bool):
        self._edit_field("advanced_timeout_during_generation", value)

    ######

    def _check_editable_and_mark_dirty(self):
        """Checks if _nested_autosave_level is not 0 and marks at least one value was changed."""
        assert self._nested_autosave_level != 0, "Attempting to edit an Options, but it wasn't made editable"
        self._is_dirty = True

    def _edit_field(self, field_name: str, new_value):
        current_value = getattr(self, field_name)
        if current_value != new_value:
            self._check_editable_and_mark_dirty()
            self._set_field(field_name, new_value)



import asyncio
import os
from pathlib import Path

import aiohttp

from randovania import VERSION

_ROOT_FOLDER = Path(__file__).parents[1]
API_URL = 'https://api.bintray.com'
REPOSITORY_USER = 'randovania'
REPOSITORY = 'randovania'
PACKAGE = 'randovania'


async def main():
    release_files = [f for f in _ROOT_FOLDER.joinpath("dist").glob("randovania-*.*")]
    print(f"Found {len(release_files)} files at {_ROOT_FOLDER.joinpath('dist')}:")
    for f in release_files:
        print(f"* {f}")

    print("\nLooking for a valid file...")
    zip_file = None
    for f in release_files:
        if f.name.endswith(".zip") or f.name.endswith(".tar.gz") or f.name.endswith(".7z"):
            zip_file = f
            print(f"Upload release at {zip_file}")
            break
        else:
            print(f"* {f.name} is not a .zip or .tar.gz or .7z")

    if zip_file is None:
        raise RuntimeError(f"No valid release file found.")

    bintray_user = os.environ["BINTRAY_USER"]
    bintray_password = os.environ["BINTRAY_API_KEY"]

    upload_url = f"{API_URL}/content/{REPOSITORY_USER}/{REPOSITORY}/{PACKAGE}/{VERSION}/{zip_file.name};publish=1"
    async with aiohttp.ClientSession(auth=aiohttp.BasicAuth(bintray_user, bintray_password)) as session:
        async with session.put(upload_url, data=zip_file.open("rb")) as response:
            print(await response.text())


if __name__ == '__main__':
    asyncio.run(main())



import dataclasses
from random import Random
from unittest.mock import MagicMock, patch, call, ANY

import pytest

from randovania.game_description import data_reader
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    HintDarkTemple
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.games.game import RandovaniaGame
from randovania.generator import base_patches_factory
from randovania.layout.elevators import LayoutElevators
from randovania.layout.translator_configuration import LayoutTranslatorRequirement


@pytest.mark.parametrize("skip_final_bosses", [False, True])
def test_add_elevator_connections_to_patches_vanilla(echoes_game_data,
                                                     skip_final_bosses: bool,
                                                     default_layout_configuration):
    # Setup
    game = data_reader.decode_data(echoes_game_data)
    expected = dataclasses.replace(game.create_game_patches())
    if skip_final_bosses:
        expected.elevator_connection[136970379] = AreaLocation(1006255871, 1393588666)

    # Run
    result = base_patches_factory.add_elevator_connections_to_patches(
        dataclasses.replace(default_layout_configuration, skip_final_bosses=skip_final_bosses),
        Random(0),
        game.create_game_patches())

    # Assert
    assert result == expected


@pytest.mark.parametrize("skip_final_bosses", [False, True])
def test_add_elevator_connections_to_patches_random(echoes_game_data,
                                                    skip_final_bosses: bool,
                                                    default_layout_configuration):
    # Setup
    game = data_reader.decode_data(echoes_game_data)
    layout_configuration = dataclasses.replace(default_layout_configuration,
                                               elevators=LayoutElevators.TWO_WAY_RANDOMIZED,
                                               skip_final_bosses=skip_final_bosses)
    expected = dataclasses.replace(game.create_game_patches(),
                                   elevator_connection={
                                       589851: AreaLocation(1039999561, 1868895730),
                                       1572998: AreaLocation(1039999561, 3479543630),
                                       1966093: AreaLocation(2252328306, 408633584),
                                       2097251: AreaLocation(1119434212, 3331021649),
                                       136970379: (AreaLocation(1006255871, 1393588666)
                                                   if skip_final_bosses else AreaLocation(2252328306, 2068511343)),
                                       3342446: AreaLocation(1039999561, 3205424168),
                                       3538975: AreaLocation(1119434212, 2806956034),
                                       152: AreaLocation(1006255871, 2889020216),
                                       393260: AreaLocation(464164546, 3145160350),
                                       524321: AreaLocation(464164546, 900285955),
                                       589949: AreaLocation(1006255871, 2278776548),
                                       122: AreaLocation(464164546, 3528156989),
                                       1245307: AreaLocation(1006255871, 1345979968),
                                       2949235: AreaLocation(1006255871, 1287880522),
                                       129: AreaLocation(1006255871, 2918020398),
                                       2162826: AreaLocation(1006255871, 1660916974),
                                       4522032: AreaLocation(1006255871, 3455543403),
                                       38: AreaLocation(1119434212, 1473133138),
                                       1245332: AreaLocation(2252328306, 2399252740),
                                       1638535: AreaLocation(2252328306, 2556480432),
                                       204865660: AreaLocation(464164546, 1564082177),
                                       4260106: AreaLocation(464164546, 3136899603),
                                   })

    # Run
    result = base_patches_factory.add_elevator_connections_to_patches(layout_configuration,
                                                                      Random(0),
                                                                      game.create_game_patches(),
                                                                      )

    # Assert
    assert result == expected


def test_gate_assignment_for_configuration_all_emerald(echoes_resource_database):
    # Setup
    emerald = find_resource_info_with_long_name(echoes_resource_database.item, "Emerald Translator")
    indices = [1, 15, 23]

    configuration = MagicMock()
    configuration.translator_configuration.translator_requirement = {
        TranslatorGate(index): LayoutTranslatorRequirement.EMERALD
        for index in indices
    }

    rng = MagicMock()

    # Run
    results = base_patches_factory.gate_assignment_for_configuration(configuration, echoes_resource_database, rng)

    # Assert
    assert results == {
        TranslatorGate(index): emerald
        for index in indices
    }


def test_gate_assignment_for_configuration_all_random(echoes_resource_database):
    # Setup
    violet = find_resource_info_with_long_name(echoes_resource_database.item, "Violet Translator")
    emerald = find_resource_info_with_long_name(echoes_resource_database.item, "Emerald Translator")

    configuration = MagicMock()
    configuration.translator_configuration.translator_requirement = {
        TranslatorGate(index): LayoutTranslatorRequirement.RANDOM
        for index in [1, 15, 23]
    }

    rng = MagicMock()
    rng.choice.side_effect = [
        LayoutTranslatorRequirement.EMERALD,
        LayoutTranslatorRequirement.VIOLET,
        LayoutTranslatorRequirement.EMERALD,
    ]

    # Run
    results = base_patches_factory.gate_assignment_for_configuration(configuration, echoes_resource_database, rng)

    # Assert
    assert results == {
        TranslatorGate(1): emerald,
        TranslatorGate(15): violet,
        TranslatorGate(23): emerald,
    }


@pytest.mark.parametrize("is_multiworld", [False, True])
def test_add_default_hints_to_patches(echoes_game_description, empty_patches, is_multiworld):
    # Setup
    rng = MagicMock()

    def _light_suit_location_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.LIGHT_SUIT_LOCATION,
                                                     HintItemPrecision.DETAILED, include_owner=False),
                    PickupIndex(number))

    def _guardian_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.GUARDIAN,
                                                     HintItemPrecision.DETAILED, include_owner=False),
                    PickupIndex(number))

    def _keybearer_hint(number: int):
        return Hint(HintType.LOCATION, PrecisionPair(HintLocationPrecision.KEYBEARER,
                                                     HintItemPrecision.BROAD_CATEGORY,
                                                     include_owner=True),
                    PickupIndex(number))

    expected = {
        # Keybearer
        LogbookAsset(0xE3B417BF): _keybearer_hint(11),
        LogbookAsset(0x65206511): _keybearer_hint(15),
        LogbookAsset(0x28E8C41A): _keybearer_hint(19),
        # Agon
        LogbookAsset(0x150E8DB8): _keybearer_hint(45),
        LogbookAsset(0xDE525E1D): _keybearer_hint(53),
        # Torvus
        LogbookAsset(0x58C62CB3): _keybearer_hint(68),
        LogbookAsset(0x939AFF16): _keybearer_hint(91),
        # Sanctuary
        LogbookAsset(0x62CC4DC3): _keybearer_hint(117),
        LogbookAsset(0xA9909E66): _keybearer_hint(106),

        # Locations with hints
        LogbookAsset(1041207119): _light_suit_location_hint(24),
        LogbookAsset(4115881194): _guardian_hint(43),
        LogbookAsset(1948976790): _guardian_hint(79),
        LogbookAsset(3212301619): _guardian_hint(115),

        # Dark Temple hints
        LogbookAsset(67497535): Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.AGON_WASTES),
        LogbookAsset(4072633400): Hint(HintType.RED_TEMPLE_KEY_SET, None, dark_temple=HintDarkTemple.TORVUS_BOG),
        LogbookAsset(0x82919C91): Hint(HintType.RED_TEMPLE_KEY_SET, None,
                                       dark_temple=HintDarkTemple.SANCTUARY_FORTRESS),

        # Jokes
        LogbookAsset(0x49CD4F34): Hint(HintType.JOKE, None),
        LogbookAsset(0x9F94AC29): Hint(HintType.JOKE, None),
    }

    # Run
    result = base_patches_factory.add_echoes_default_hints_to_patches(
        rng, empty_patches, echoes_game_description.world_list, num_joke=2, is_multiworld=is_multiworld)

    # Assert
    rng.shuffle.assert_has_calls([call(ANY), call(ANY)])
    assert result.hints == expected


@patch("randovania.generator.base_patches_factory.add_echoes_default_hints_to_patches", autospec=True)
@patch("randovania.generator.base_patches_factory.starting_location_for_configuration", autospec=True)
@patch("randovania.generator.base_patches_factory.gate_assignment_for_configuration", autospec=True)
@patch("randovania.generator.base_patches_factory.add_elevator_connections_to_patches", autospec=True)
def test_create_base_patches(mock_add_elevator_connections_to_patches: MagicMock,
                             mock_gate_assignment_for_configuration: MagicMock,
                             mock_starting_location_for_config: MagicMock,
                             mock_add_default_hints_to_patches: MagicMock,
                             mocker,
                             ):
    # Setup
    rng = MagicMock()
    game = MagicMock()
    layout_configuration = MagicMock()
    layout_configuration.game = RandovaniaGame.PRIME2
    mock_replace: MagicMock = mocker.patch("dataclasses.replace")
    is_multiworld = MagicMock()

    patches = ([
        game.create_game_patches.return_value,
        mock_replace.return_value,
        mock_add_elevator_connections_to_patches.return_value,
    ])
    patches.append(patches[-1].assign_gate_assignment.return_value)
    patches.append(patches[-1].assign_starting_location.return_value)

    # Run
    result = base_patches_factory.create_base_patches(layout_configuration, rng, game, is_multiworld)

    # Assert
    game.create_game_patches.assert_called_once_with()
    mock_replace.assert_called_once_with(game.create_game_patches.return_value, game_specific=ANY)
    mock_add_elevator_connections_to_patches.assert_called_once_with(layout_configuration, rng, patches[1])

    # Gate Assignment
    mock_gate_assignment_for_configuration.assert_called_once_with(layout_configuration, game.resource_database, rng)
    patches[2].assign_gate_assignment.assert_called_once_with(mock_gate_assignment_for_configuration.return_value)

    # Starting Location
    mock_starting_location_for_config.assert_called_once_with(layout_configuration, game, rng)
    patches[3].assign_starting_location.assert_called_once_with(mock_starting_location_for_config.return_value)

    # Hints
    mock_add_default_hints_to_patches.assert_called_once_with(rng, patches[4], game.world_list, num_joke=2,
                                                              is_multiworld=is_multiworld)

    assert result is mock_add_default_hints_to_patches.return_value



from pathlib import Path
from unittest.mock import patch, MagicMock

import pytest

from randovania.interface_common import simplified_patcher
from randovania.interface_common.options import Options
from randovania.layout.permalink import Permalink


@pytest.mark.parametrize("games_path_exist", [False, True])
@pytest.mark.parametrize("backup_path_exist", [False, True])
def test_delete_files_location(tmpdir,
                               games_path_exist: bool,
                               backup_path_exist: bool,
                               ):
    # Setup
    data_dir = Path(str(tmpdir.join("user_data_dir")))
    options = Options(data_dir)

    game_files = tmpdir.join("user_data_dir", "extracted_game")
    if games_path_exist:
        game_files.ensure_dir()
        game_files.join("random.txt").write_text("yay", "utf-8")

    backup_files = tmpdir.join("user_data_dir", "backup")
    if backup_path_exist:
        backup_files.ensure_dir()
        backup_files.join("random.txt").write_text("yay", "utf-8")

    # Run
    simplified_patcher.delete_files_location(options)

    # Assert
    assert not game_files.exists()
    assert not backup_files.exists()


@patch("randovania.interface_common.simplified_patcher.iso_packager.unpack_iso", autospec=True)
@patch("randovania.interface_common.simplified_patcher.delete_files_location", autospec=True)
def test_unpack_iso(mock_delete_files_location: MagicMock,
                    mock_unpack_iso: MagicMock,
                    ):
    # Setup
    input_iso: Path = MagicMock()
    options: Options = MagicMock()
    progress_update = MagicMock()

    # Run
    simplified_patcher.unpack_iso(input_iso, options, progress_update)

    # Assert
    mock_delete_files_location.assert_called_once_with(options)
    mock_unpack_iso.assert_called_once_with(
        iso=input_iso,
        game_files_path=options.game_files_path,
        progress_update=progress_update,
    )


@patch("randovania.interface_common.simplified_patcher.ConstantPercentageCallback",
       autospec=False)  # TODO: pytest-qt bug
@patch("randovania.interface_common.echoes.generate_layout", autospec=True)
def test_generate_layout(mock_generate_layout: MagicMock,
                         mock_constant_percentage_callback: MagicMock,
                         ):
    # Setup
    options: Options = MagicMock()
    permalink: Permalink = MagicMock()
    progress_update = MagicMock()

    # Run
    simplified_patcher.generate_layout(options, permalink, progress_update)

    # Assert
    mock_constant_percentage_callback.assert_called_once_with(progress_update, -1)
    mock_generate_layout.assert_called_once_with(
        permalink=permalink,
        status_update=mock_constant_percentage_callback.return_value,
        validate_after_generation=options.advanced_validate_seed_after,
        timeout_during_generation=options.advanced_timeout_during_generation,
    )


@patch("randovania.games.prime.claris_randomizer.apply_layout", autospec=True)
@patch("randovania.interface_common.simplified_patcher.patch_game_name_and_id", autospec=True)
def test_apply_layout(mock_patch_game_name_and_id: MagicMock,
                      mock_claris_apply_layout: MagicMock,
                      empty_patches
                      ):
    # Setup
    layout = MagicMock()
    progress_update = MagicMock()
    options: Options = MagicMock()
    players_config = MagicMock()

    # Run
    simplified_patcher.apply_layout(layout, options, players_config, progress_update)

    # Assert
    mock_patch_game_name_and_id.assert_called_once_with(
        options.game_files_path, "Metroid Prime 2: Randomizer - {}".format(layout.shareable_hash)
    )
    mock_claris_apply_layout.assert_called_once_with(
        description=layout,
        players_config=players_config,
        cosmetic_patches=options.cosmetic_patches,
        game_root=options.game_files_path,
        backup_files_path=options.backup_files_path,
        progress_update=progress_update,
    )


@patch("randovania.interface_common.simplified_patcher.patch_game_name_and_id", autospec=True)
@patch("randovania.games.prime.claris_randomizer.apply_patcher_file", autospec=True)
def test_apply_patcher_file(mock_apply_patcher_file: MagicMock,
                            mock_patch_game_name_and_id: MagicMock,
                            ):
    # Setup
    patcher_data = MagicMock()
    game_specific = MagicMock()
    progress_update = MagicMock()
    options: Options = MagicMock()
    shareable_hash = "Some Magical Hash"

    # Run
    simplified_patcher.apply_patcher_file(patcher_data, game_specific, shareable_hash, options, progress_update)

    # Assert
    mock_patch_game_name_and_id.assert_called_once_with(options.game_files_path,
                                                        f"Metroid Prime 2: Randomizer - {shareable_hash}")
    mock_apply_patcher_file.assert_called_once_with(
        game_root=options.game_files_path,
        backup_files_path=options.backup_files_path,
        patcher_data=patcher_data,
        game_specific=game_specific,
        progress_update=progress_update,
    )



from typing import NamedTuple, Optional

from randovania.game_description.resources.resource_type import ResourceType


class ItemResourceInfo(NamedTuple):
    index: int
    long_name: str
    short_name: str
    max_capacity: int
    extra: Optional[int]

    def __str__(self):
        return self.long_name

    @property
    def resource_type(self) -> ResourceType:
        return ResourceType.ITEM
